/*
 * Copyright (c) 2006 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintStream;
import java.io.Reader;
import java.io.StringWriter;
import java.net.Proxy;
import java.net.URL;
import java.net.URLConnection;
import java.util.HashMap;
import java.util.Map;
import java.util.Random;

/**
 * A utility class used for downloading files from the Internet for parsing.
 * This class provides a built-in disk cache to avoid redownload of the same
 * URLs after a crash or after fixing a bug.
 */
public class CachingDownloader {

	private final File cachePath;
	private final Proxy proxy;
	private final String cookies, useragent, forwardedFor;
	private final long delay;
	private final Random rnd = new Random();
	private final int blacklistedSize;
	private final Map<String, int[]> cacheNameLocks = new HashMap<String, int[]>();
	private PrintStream debugStream = System.err;

	/**
	 * Create a new caching downloader that stores its cache in the given
	 * directory.
	 * 
	 * @param cachePath
	 *            directory to store the cached files into
	 */
	public CachingDownloader(File cachePath) {
		this(cachePath, null, null);
	}

	/**
	 * Create a new caching downloader with support for custom user agents and
	 * cookies
	 * 
	 * @param cachePath
	 *            directory to store the cached files into
	 * @param cookies
	 *            Value for the <tt>Cookie</tt> header
	 * @param useragent
	 *            Value for the <tt>User-Agent</tt> header
	 */
	public CachingDownloader(File cachePath, String cookies, String useragent) {
		this(cachePath, Proxy.NO_PROXY, cookies, useragent, null, 0, -1);
	}

	/**
	 * Create a new caching downloader with all supported options.
	 * 
	 * @param cachePath
	 *            Directory to store the cached files into
	 * @param proxy
	 *            Proxy to use for downloading
	 * @param cookies
	 *            Value for the <tt>Cookie</tt> header
	 * @param useragent
	 *            Value for the <tt>User-Agent</tt> header
	 * @param forwardedFor
	 *            Value for the <tt>X-Forwarded-For</tt> header. Every
	 *            <tt>*</tt> will be replaced by a random number between 0 and
	 *            255 for each request
	 * @param delay
	 *            Delay to wait before each download (useful if the target site
	 *            blocks excess downloaders)
	 * @param blacklistedSize
	 *            Size of a proxy error page. If the response has the given
	 *            size, the download will be repeated. Only needed if the proxy
	 *            uses status code 200 for its error pages.
	 */
	public CachingDownloader(File cachePath, Proxy proxy, String cookies, String useragent, String forwardedFor, long delay, int blacklistedSize) {
		this.cachePath = cachePath;
		this.cookies = cookies;
		this.useragent = useragent;
		this.forwardedFor = forwardedFor;
		this.delay = delay;
		this.proxy = proxy;
		this.blacklistedSize = blacklistedSize;
	}

	/**
	 * Download the given file.
	 * 
	 * @param url
	 *            URL of the file
	 * @param cacheName
	 *            Name to use for the file in the cache (must be a valid file
	 *            name)
	 * @return An input stream to read from the file
	 */
	public InputStream download(String url, String cacheName) throws IOException {
		return download(url, null, cacheName);
	}

	/**
	 * Download the given file using user defined POST data.
	 * 
	 * @param url
	 *            URL of the file
	 * @param postdata
	 *            POST data to send to the URL, or <code>null</code> to not use
	 *            any POST data
	 * @param cacheName
	 *            Name to use for the file in the cache (must be a valid file
	 *            name)
	 * @return An input stream to read from the file
	 */
	public InputStream download(String url, String postdata, String cacheName) throws IOException {
		File cacheFile = new File(cachePath, cacheName);
		int[] lock;
		synchronized (cacheNameLocks) {
			lock = cacheNameLocks.get(cacheName);
			if (lock == null) {
				lock = new int[] { 0 };
				cacheNameLocks.put(cacheName, lock);
			}
			lock[0]++;
		}
		try {
			synchronized (lock) {
				if (!cacheFile.exists()) {
					download(url, postdata, cacheFile);
				}
			}
		} finally {
			synchronized (cacheNameLocks) {
				lock[0]--;
				if (lock[0] == 0) {
					cacheNameLocks.remove(cacheName);
				}
			}
		}
		return new FileInputStream(cacheFile);
	}

	/**
	 * Download the given file using user defined POST data into the given cache
	 * file. This (private) method does not enforce any lockin, so you have to
	 * make sure yourself not to call it twice for the same file.
	 * 
	 * @param url
	 *            URL of the file
	 * @param postdata
	 *            POST data to send to the URL, or <code>null</code> to not use
	 *            any POST data
	 * @param cacheFile
	 *            File to use to store the download to
	 */
	private void download(String url, String postdata, File cacheFile) throws IOException {
		if (delay > 0) {
			debugStream.println("Waiting for " + url + " " + postdata);
			try {
				Thread.sleep(delay);
			} catch (InterruptedException ex) {
			}
		}
		debugStream.println("Downloading " + url + " " + postdata);
		URLConnection conn = new URL(url).openConnection(proxy);
		if (cookies != null) {
			conn.setRequestProperty("Cookie", cookies);
		}
		if (useragent != null) {
			conn.setRequestProperty("User-Agent", useragent);
		}
		if (forwardedFor != null) {
			String ff = forwardedFor;
			synchronized (rnd) {
				while (ff.indexOf('*') != -1) {
					int pos = ff.indexOf('*');
					ff = ff.substring(0, pos) + rnd.nextInt(256) + ff.substring(pos + 1);
				}
			}
			conn.setRequestProperty("X-Forwarded-For", ff);
		}
		if (postdata != null) {
			conn.setDoOutput(true);
			OutputStreamWriter wr = new OutputStreamWriter(conn.getOutputStream(), "ISO-8859-1");
			wr.write(postdata);
			wr.flush();
		}
		InputStream in = conn.getInputStream();
		FileOutputStream out = new FileOutputStream(cacheFile);
		StreamForwarder.forward(in, out);
		if (blacklistedSize >= 0 && cacheFile.length() == blacklistedSize) {
			debugStream.println("------ Blacklisted size, retrying...");
			cacheFile.delete();
			download(url, postdata, cacheFile);
		}
	}

	/**
	 * Load an {@link InputStream} completely into a {@link String}.
	 * 
	 * @param in
	 *            Stream to read from
	 * @param encoding
	 *            Encoding to use
	 * @return complete content of the stream
	 */
	public static String loadStream(InputStream in, String encoding) throws IOException {
		Reader r = new InputStreamReader(in, encoding);
		StringWriter sw = new StringWriter();
		StreamForwarder.forward(r, sw);
		return sw.toString();
	}
	
	/**
	 * Set the debug stream where status information is written to
	 * ({@link System#err} by default).
	 * 
	 * @param debugStream
	 *            New debug stream to use
	 */
	public void setDebugStream(PrintStream debugStream) {
		this.debugStream = debugStream;
	}
}
