/*
 * Copyright (c) 2006 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util;

import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

/**
 * This exception is thrown when an expectation fails. It tries to show useful
 * information (like actual and expected values) in its message.
 */
public class ExpectException extends RuntimeException {

	private static final long serialVersionUID = 4350118922979233122L;

	/**
	 * Create a new Expect exception
	 * 
	 * @param actual
	 *            Actual value
	 * @param expected
	 *            Expected value
	 */
	public ExpectException(String actual, String expected) {
		super("Unexpected string\nActual:   '" + actual + "'\nExpected: '" + expected + "'");
	}

	private ExpectException(String actual, Pattern pattern) {
		super("Nonmatching regex\n\nActual: '" + actual + "'\nRegex:  '" + pattern.pattern() + "'");
	}

	private ExpectException(String actual, Pattern pattern, String debugActual, String debugPattern) {
		super("Nonmatching regex\n\nActual: '" + actual + "'\nRegex:  '" + pattern.pattern() + "'\n\nActual part: '" + debugActual + "'\nRegex part:  '" + debugPattern + "'");
	}

	private ExpectException(String actual, Pattern pattern, Exception cause) {
		super("Nonmatching regex\n\nActual: '" + actual + "'\nRegex:  '" + pattern.pattern() + "'\n\nNo debug information available", cause);
	}

	/**
	 * Create a new expect exception from a regular expression.
	 * 
	 * @param actual
	 *            Actual value
	 * @param pattern
	 *            Pattern that should match
	 */
	public static ExpectException fromRegex(final String actual, final Pattern pattern) {
		try {
			String p = pattern.pattern();
			Pattern pt = pattern;
			while (!pt.matcher(actual).matches() && p.indexOf(">{0}") != -1) {
				p = p.substring(0, p.lastIndexOf(">{0}")) + "(.*)";
				pt = PatternCache.compile(p);
			}
			if (!pattern.pattern().equals(p)) {
				String restpattern = pattern.pattern().substring(p.length() - 4 + 4);
				Matcher m = pt.matcher(actual);
				if (!m.matches())
					throw new RuntimeException();
				String restactual = m.group(m.groupCount());
				return new ExpectException(actual, pattern, restactual, restpattern);
			}
			return new ExpectException(actual, pattern);
		} catch (PatternSyntaxException ex) {
			return new ExpectException(actual, pattern, ex);
		}
	}
}