/*
 * Copyright (c) 2006 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util;

import java.util.ArrayList;
import java.util.List;
import java.io.BufferedReader;
import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.UnsupportedEncodingException;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * A fairly simple reader class, that can "expect" constant text and regular
 * expressions, while still leaving full control to the programmer. This parser
 * is fully deterministic; therefore, using it on more complex documents can be
 * quite frustrating. Basically, this is what I used before I wrote
 * JMatchParser. It is included because I still have old projects using it, and
 * because it might be useful for simple projects.
 */
public class ExpectReader {

	private final BufferedReader br;

	/**
	 * Create a new expect reader reading from the given {@link BufferedReader}.
	 */
	public ExpectReader(BufferedReader br) {
		this.br = br;
	}

	/**
	 * Create a new expect reader reading from the given {@link Reader}.
	 */
	public ExpectReader(Reader r) {
		this(new BufferedReader(r));
	}

	/**
	 * Create a new expect reader reading from the given {@link InputStream}.
	 */
	public ExpectReader(InputStream in, String charsetName) throws UnsupportedEncodingException {
		this(new InputStreamReader(in, charsetName));
	}

	// generic methods

	/**
	 * Return the reader used.
	 */
	public BufferedReader getReader() {
		return br;
	}

	/**
	 * Read a line from the reader, and return it. Throw an exception if eof is
	 * reached before reading the line.
	 */
	public String ensureReadLine() throws IOException {
		String l = readLine();
		if (l == null)
			throw new EOFException();
		return l;
	}

	/**
	 * Read a line like {@link #ensureReadLine()}, but trim the response and
	 * skip empty lines.
	 */
	public String ensureReadTrimmedLine() throws IOException {
		String l = ensureReadLine().trim();
		while (l.length() == 0) {
			l = ensureReadLine().trim();
		}
		return l;
	}

	/**
	 * Read a line.
	 * 
	 * @see BufferedReader#readLine()
	 */
	public String readLine() throws IOException {
		return br.readLine();
	}

	/**
	 * Ensure that the end of file has been reached.
	 */
	public void ensureEndOfFile() throws IOException {
		if (br.readLine() != null)
			throw new IOException("Not at end of file");
	}

	/**
	 * Close the underlying reader.
	 */
	public void close() throws IOException {
		br.close();
	}

	// string matches

	/**
	 * Expect one ore more lines.
	 * 
	 * @see #expectLine(String)
	 */
	public void expect(String... lines) throws IOException {
		for (String l : lines) {
			expectLine(l);
		}
	}

	/**
	 * Expect one ore more trimmed lines.
	 * 
	 * @see #expectTrimmedLine(String)
	 */
	public void expectTrimmed(String... lines) throws IOException {
		for (String l : lines) {
			expectTrimmedLine(l);
		}
	}

	/**
	 * Expect a line. If the {@link Reader}'s next line does not match the line
	 * given, an exception is thrown.
	 */
	public void expectLine(String line) throws IOException {
		ensureStringMatch(ensureReadLine(), line);
	}

	/**
	 * Expect a trimmed. If the {@link Reader}'s next line does not match the
	 * line given, after trimming, an exception is thrown.
	 * 
	 * For symmetry reasons, calling this method with an empty string is a
	 * no-op.
	 */
	public void expectTrimmedLine(String line) throws IOException {
		if (line.trim().length() == 0)
			return;
		ensureStringMatch(ensureReadTrimmedLine(), line.trim());
	}

	/**
	 * Ensure that two strings are the same.
	 * 
	 * @param actual
	 *            The actual string
	 * @param expected
	 *            The expected string
	 */
	public static void ensureStringMatch(String actual, String expected) {
		if (!expected.equals(actual)) {
			throw new ExpectException(actual, expected);
		}
	}

	// regex matches

	/**
	 * Expect a line that matches a regex. A {@link Matcher} to read the groups
	 * is returned.
	 */
	public Matcher expectRegex(String regex) throws IOException {
		return ensureRegexMatch(ensureReadLine(), regex);
	}

	/**
	 * Expect a line that matches a regex. A {@link Matcher} to read the groups
	 * is returned.
	 */
	public Matcher expectRegex(Pattern pattern) throws IOException {
		return ensureRegexMatch(ensureReadLine(), pattern);
	}

	/**
	 * Expect a trimmed line that matches a regex. A {@link Matcher} to read the
	 * groups is returned.
	 * 
	 * @see #ensureReadTrimmedLine()
	 */
	public Matcher expectTrimmedRegex(String regex) throws IOException {
		return ensureRegexMatch(ensureReadTrimmedLine(), regex);
	}

	/**
	 * Expect a trimmed line that matches a regex. A {@link Matcher} to read the
	 * groups is returned.
	 * 
	 * @see #ensureReadTrimmedLine()
	 */
	public Matcher expectTrimmedRegex(Pattern pattern) throws IOException {
		return ensureRegexMatch(ensureReadTrimmedLine(), pattern);
	}

	/**
	 * Ensure that a string matches a pattern and returns a {@link Matcher}.
	 * 
	 * @param actual
	 *            The actual string
	 * @param pattern
	 *            The pattern
	 */
	public static Matcher ensureRegexMatch(String actual, String pattern) {
		return ensureRegexMatch(actual, PatternCache.compile(pattern));
	}

	/**
	 * Ensure that a string matches a pattern and returns a {@link Matcher}.
	 * 
	 * @param actual
	 *            The actual string
	 * @param pattern
	 *            The pattern
	 */
	public static Matcher ensureRegexMatch(String actual, Pattern pattern) {
		Matcher m = regexMatch(actual, pattern);
		if (m == null)
			throw ExpectException.fromRegex(actual, pattern);
		return m;
	}

	/**
	 * Try to match a string against pattern and returns a {@link Matcher} or
	 * <code>null</code>.
	 * 
	 * @param actual
	 *            The actual string
	 * @param pattern
	 *            The pattern
	 */
	public static Matcher regexMatch(String actual, String pattern) {
		return regexMatch(actual, PatternCache.compile(pattern));
	}

	/**
	 * Try to match a string against pattern and returns a {@link Matcher} or
	 * <code>null</code>.
	 * 
	 * @param actual
	 *            The actual string
	 * @param pattern
	 *            The pattern
	 */
	public static Matcher regexMatch(String actual, Pattern pattern) {
		Matcher m = pattern.matcher(actual);
		if (m.matches()) {
			return m;
		}
		return null;
	}

	// multi regex match

	/**
	 * Perform a multi regex match.
	 * 
	 * @see #ensureMultiRegexMatch(String, Pattern, Pattern, int)
	 */
	public static List<Matcher> ensureMultiRegexMatch(String actual, String regex, String terminalRegex, int restGroup) {
		return ensureMultiRegexMatch(actual, PatternCache.compile(regex),
				terminalRegex == null ? null : PatternCache.compile(terminalRegex),
						restGroup);
	}

	/**
	 * Perform a multi regex match.
	 * 
	 * <p>
	 * If terminalPattern is non-<code>null</code>: Match actual against the
	 * terminal pattern. If it matches, the match is finished. If not, it is
	 * matched against the pattern and added to the result list (non-match is an
	 * error). The restGroup'th group of the matcher is then used as the actual
	 * value for the next round.
	 * 
	 * <p>
	 * If terminalPattern is <code>null</code>: Terminate when it does not match
	 * any longer. Rest as above.
	 */
	public static List<Matcher> ensureMultiRegexMatch(String actual, Pattern pattern, Pattern terminalPattern, int restGroup) {
		List<Matcher> result = new ArrayList<Matcher>();
		while (terminalPattern == null || !terminalPattern.matcher(actual).matches()) {
			Matcher m = pattern.matcher(actual);
			if (!m.matches()) {
				if (terminalPattern == null)
					return result;
				throw ExpectException.fromRegex(actual, pattern);
			}
			result.add(m);
			actual = m.group(restGroup);
		}
		return result;
	}
}