/*
 * Copyright (c) 2006 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util;

import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * A class that can "expect" parts at the start or end of a string, and get/set
 * the remaining part (rest). In most cases, {@link ExpectReader} is more
 * flexible, but sometimes this class might be more handy.
 */
public class ExpectString {

	private String rest;
	private int offset;

	/**
	 * Create a new expect string
	 * 
	 * @param rest
	 *            The rest still to be matched
	 */
	public ExpectString(String rest) {
		this.rest = rest;
		this.offset = 0;
	}

	// generic methods

	private void cleanup() {
		if (offset != 0) {
			rest = rest.substring(offset);
			offset = 0;
		}
	}

	/**
	 * Return the rest that has not yet been parsed.
	 */
	public String getRest() {
		return rest.substring(offset);
	}

	/**
	 * Change the rest that has not yet been parsed, and return the old rest.
	 */
	public String changeRest(String newRest) {
		String oldRest = rest.substring(offset);
		rest = newRest;
		offset = 0;
		return oldRest;
	}

	/**
	 * Expect some variable text before some fixed string.
	 * 
	 * @param s
	 *            the fixed string
	 * @return the variable text
	 */
	public String expectStringBeforeString(String s) {
		int pos = rest.indexOf(s, offset);
		if (pos == -1)
			throw new ExpectException(getRest(), "'...'" + s);
		String result = rest.substring(offset, pos);
		offset = pos + s.length();
		return result;
	}

	/**
	 * Expect some fixed string followed by some variable text at the end of the
	 * rest.
	 * 
	 * @param s
	 *            the fixed string
	 * @return the variable text
	 */
	public String expectStringAfterString(String s) {
		cleanup();
		int pos = rest.lastIndexOf(s);
		if (pos == -1)
			throw new ExpectException(rest, s + "'...'");
		String result = rest.substring(pos + s.length());
		rest = rest.substring(0, pos);
		return result;
	}

	/**
	 * Expect a fixed string at the beginning.
	 */
	public void expect(String start) {
		if (!rest.startsWith(start, offset)) {
			throw new ExpectException(getRest(), start);
		}
		offset += start.length();
	}

	/**
	 * Expect a fixed string at the end.
	 */
	public void expectEnd(String end) {
		if (!rest.endsWith(end)) {
			throw new ExpectException(getRest(), end);
		}
		rest = rest.substring(offset, rest.length() - end.length());
		offset = 0;
	}

	// regex matches

	/**
	 * Expect a regex at the begin of the string and return a matcher for it.
	 * The matcher will have an additional group containing the rest.
	 */
	public Matcher expectRegex(String regex) {
		return expectRegex(regex + "(.*)", -1);
	}

	/**
	 * Expect a regex that matches the complete string. The keptGroup will
	 * remain in the rest.
	 */
	public Matcher expectRegex(String pattern, int keptGroup) {
		return expectRegex(PatternCache.compile(pattern), keptGroup);
	}

	/**
	 * Expect a regex that matches the complete string. The keptGroup will
	 * remain in the rest.
	 */
	public Matcher expectRegex(Pattern pattern, int keptGroup) {
		cleanup();
		Matcher m = pattern.matcher(rest);
		if (!m.matches()) {
			throw ExpectException.fromRegex(rest, pattern);
		}
		if (keptGroup == -1)
			keptGroup = m.groupCount();
		rest = m.group(keptGroup);
		return m;
	}

	// multi regex match

	/**
	 * Match the given regex more than once against the rest.
	 * 
	 * @see ExpectReader#ensureMultiRegexMatch(String, Pattern, Pattern, int)
	 */
	public List<Matcher> expectMultiRegexMatch(String regex, int restGroup) {
		return expectMultiRegexMatch(PatternCache.compile(regex),
						restGroup);
	}

	/**
	 * Match the given regex more than once against the rest.
	 * 
	 * @see ExpectReader#ensureMultiRegexMatch(String, Pattern, Pattern, int)
	 */
	public List<Matcher> expectMultiRegexMatch(Pattern pattern, int restGroup) {
		cleanup();
		List<Matcher> lm = ExpectReader.ensureMultiRegexMatch(rest, pattern, null, restGroup);
		if (lm.size() > 0) {
			rest = lm.get(lm.size() - 1).group(restGroup);
		}
		return lm;
	}
}
