/*
 * Copyright (c) 2010 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Reader;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * A {@link Reader} that reads from a {@link BufferedReader} and counts the
 * lines read. At the end of the output, there is a marker followed by the
 * actual line count. If the line count is wrong, an exception is thrown, else
 * the line count and everything after it is ignored.
 * 
 * As a convenience, when closing a not fully consumed reader, the remaining
 * lines are read and the line count is verified. Therefore, to verify the line
 * count without reading the contents, it is sufficient to close the reader.
 */
public class LineCountVerifyReader extends Reader {

	private final BufferedReader br;
	private String currentLine = null;
	private final Pattern lineCountMarkerPattern;
	private final Pattern lineCountPattern;
	private int counter;
	private boolean counterFound;

	/**
	 * Create a new reader.
	 * 
	 * @param br
	 *            Buffered reader to read from
	 * @param lineCountMarkerPattern
	 *            Pattern of the line that marks the end of the file and the
	 *            following line count
	 * @param lineCountPattern
	 *            Pattern of the line that contains the line count as group 1.
	 *            If <code>null</code> group 1 of the previous pattern contains
	 *            the line count
	 * @param lineCountOffset
	 *            Difference between number of lines and line count printed.
	 *            Positive if the printed line count is higher and vice versa
	 */
	public LineCountVerifyReader(BufferedReader br, Pattern lineCountMarkerPattern, Pattern lineCountPattern, int lineCountOffset) {
		this.br = br;
		this.lineCountMarkerPattern = lineCountMarkerPattern;
		this.lineCountPattern = lineCountPattern;
		counter = lineCountOffset;
	}

	@Override
	public boolean ready() throws IOException {
		return currentLine != null || br.ready();
	}

	@Override
	public int read(char[] cbuf, int off, int len) throws IOException {
		ensureFillCurrentLine();
		if (currentLine == null) {
			return -1;
		}
		if (len <= currentLine.length()) {
			System.arraycopy(currentLine.toCharArray(), 0, cbuf, off, len);
			currentLine = currentLine.substring(len);
			return len;
		}
		int result = currentLine.length() + 1;
		System.arraycopy(currentLine.toCharArray(), 0, cbuf, off, currentLine.length());
		cbuf[off + currentLine.length()] = '\n';
		currentLine = null;
		return result;
	}

	@Override
	public void close() throws IOException {
		char[] buf = new char[256];
		while (read(buf) != -1)
			;
		br.close();
	}

	private void ensureFillCurrentLine() throws IOException {
		if (currentLine != null)
			return;
		currentLine = br.readLine();
		if (currentLine == null) {
			if (!counterFound)
				throw new IOException("File record count marker line missing");
			return;
		}
		Matcher m = lineCountMarkerPattern.matcher(currentLine);
		if (m.matches()) {
			if (lineCountPattern != null) {
				currentLine = br.readLine();
				m = currentLine == null ? null : lineCountPattern.matcher(currentLine);
				if (m == null || !m.matches()) {
					throw new IOException("File record count line is invalid: " + currentLine);
				}
			}
			int fileCounter = Integer.parseInt(m.group(1));
			if (fileCounter != counter)
				throw new IOException("File count incorrect: " + fileCounter + " != " + counter);
			while (br.readLine() != null)
				;
			currentLine = null;
			counterFound = true;
		}
		counter++;
	}
}