/*
 * Copyright (c) 2010 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.io.StringReader;
import java.io.StringWriter;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.stream.events.ProcessingInstruction;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.Text;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * A {@link org.w3c.dom.Document DOM} parser that annotates each
 * {@link org.w3c.dom.Node} with an approximate line/column number where it was
 * parsed.
 * 
 * Technically, this is done by parsing the document twice; therefore, this
 * class also provides a utility method to create clones of an
 * {@link InputSource}.
 */
public class LocationAwareDOMParser {

	private static final String LINE_NUMBER = LocationAwareDOMParser.class.getName() + "#LINE_NUMBER";
	private static final String COLUMN_NUMBER = LocationAwareDOMParser.class.getName() + "#COLUMN_NUMBER";

	/**
	 * Create an arbitrary number of clones of an {@link InputSource}. The
	 * original InputSource will be invalid after that, but each of the clone
	 * can be used individually to parse the same document in multiple ways.
	 * 
	 * @param input
	 *            The InputSource to clone
	 * @param count
	 *            The desired number of clones
	 * @return An array that contains the clones
	 */
	public static InputSource[] createClones(InputSource input, int count) throws IOException {
		byte[] byteBuf = null;
		InputStream byteStream = input.getByteStream();
		if (byteStream != null) {
			ByteArrayOutputStream baos = new ByteArrayOutputStream(byteStream.available());
			StreamForwarder.forward(byteStream, baos);
			byteBuf = baos.toByteArray();
		}
		String charBuf = null;
		Reader charStream = input.getCharacterStream();
		if (charStream != null) {
			StringWriter sw = new StringWriter();
			StreamForwarder.forward(charStream, sw);
			charBuf = sw.toString();
		}
		InputSource[] result = new InputSource[count];
		for (int i = 0; i < result.length; i++) {
			result[i] = new InputSource();
			if (byteBuf != null)
				result[i].setByteStream(new ByteArrayInputStream(byteBuf));
			if (charBuf != null)
				result[i].setCharacterStream(new StringReader(charBuf));
			result[i].setEncoding(input.getEncoding());
			result[i].setPublicId(input.getPublicId());
			result[i].setSystemId(input.getSystemId());
		}
		return result;
	}

	/**
	 * Parse the given {@link InputSource} to a DOM document that contains line
	 * number information, using the default {@link DocumentBuilder}.
	 */
	public static Document parse(InputSource source) throws Exception {
		return parse(DocumentBuilderFactory.newInstance().newDocumentBuilder(), source);
	}

	/**
	 * Parse the given {@link InputSource} to a DOM document that contains line
	 * number information, using a custom {@link DocumentBuilder}.
	 */
	public static Document parse(DocumentBuilder documentBuilder, InputSource source) throws Exception {
		InputSource[] sources = createClones(source, 2);
		final Document doc = documentBuilder.parse(sources[0]);
		SAXParser p = SAXParserFactory.newInstance().newSAXParser();
		p.parse(sources[1], new DefaultHandler() {

			private int nextLine = 1, nextCol = 1;
			private Locator locator;
			private Node nextNode, parentNode;

			@Override
			public void setDocumentLocator(Locator locator) {
				this.locator = locator;
			}

			@Override
			public void startDocument() throws SAXException {
				parentNode = null;
				updateLocation(doc);
				nextNode = doc.getDocumentElement();
			}

			@Override
			public void endDocument() throws SAXException {
				if (nextNode != null || parentNode != doc)
					throw new RuntimeException("Invalid state!");
			}

			@Override
			public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException {
				if (!validateNode(nextNode, uri, localName, qName, attributes))
					throw new RuntimeException("Nodes out of sync");
				updateLocation(nextNode);
				parentNode = nextNode;
				nextNode = parentNode.getFirstChild();
			}

			@Override
			public void endElement(String uri, String localName, String qName) throws SAXException {
				if (nextNode != null)
					throw new RuntimeException("Nodes out of sync");
				if (!validateNode(parentNode, uri, localName, qName, null))
					throw new RuntimeException("Nodes out of sync");
				Node currentNode = parentNode;
				updateLocation(null);
				parentNode = currentNode.getParentNode();
				nextNode = currentNode.getNextSibling();
			}

			@Override
			public void ignorableWhitespace(char[] ch, int start, int length) throws SAXException {
				updateLocation(null);
			}

			@Override
			public void characters(char[] ch, int start, int length) throws SAXException {
				if (!(nextNode instanceof Text && nextNode.getNodeValue().equals(new String(ch, start, length)))) {
					throw new RuntimeException("Nodes out of sync");
				}
				updateLocation(nextNode);
				nextNode = nextNode.getNextSibling();
			}

			@Override
			public void processingInstruction(String target, String data) throws SAXException {
				if (!(nextNode instanceof ProcessingInstruction && nextNode.getNodeName().equals(target) && nextNode.getNodeValue().equals(data))) {
					throw new RuntimeException("Nodes out of sync");
				}
				updateLocation(nextNode);
				nextNode = nextNode.getNextSibling();
			}

			private void updateLocation(Node node) {
				if (node != null) {
					node.setUserData(LINE_NUMBER, new Integer(nextLine), null);
					node.setUserData(COLUMN_NUMBER, new Integer(nextCol), null);
				}
				nextLine = locator.getLineNumber();
				nextCol = locator.getColumnNumber();
			}

			private boolean validateNode(Node node, String uri, String localName, String qName, Attributes attributes) {
				if (!(node instanceof Element))
					return false;
				if (!node.getNodeName().equals(qName))
					return false;
				else if (node.getNamespaceURI() == null && !uri.equals(""))
					return false;
				else if (node.getNamespaceURI() != null && !node.getNamespaceURI().equals(uri))
					return false;
				else if (attributes != null && node.getAttributes().getLength() != attributes.getLength())
					return false;
				else
					return true;
			}
		});
		return doc;
	}

	/**
	 * Get the line number embedded in a node.
	 * 
	 * @param node
	 *            the node
	 * @return the line number, or -1 if none was embedded
	 */
	public static int getLineNumber(Node node) {
		Integer line = (Integer) node.getUserData(LocationAwareDOMParser.LINE_NUMBER);
		return line == null ? -1 : line;
	}

	/**
	 * Get the column number embedded in a node.
	 * 
	 * @param node
	 *            the node
	 * @return the column number, or -1 if none was embedded
	 */
	public static int getColumnNumber(Node node) {
		Integer col = (Integer) node.getUserData(LocationAwareDOMParser.COLUMN_NUMBER);
		return col == null ? -1 : col;
	}
}
