/*
 * Copyright (c) 2006 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util;

import java.io.InputStream;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Queue;

/**
 * Utility class that makes it easy to download files in the background as they
 * are detected (over slow links) while still doing all of the parsing in a
 * single thread.
 * 
 * The main idea is, that whenever you encounter a file you have to parse later,
 * you add it to the prefetcher by calling the {@link #download(String, String)}
 * method. All documents added to the prefetcher are loaded in a configurable
 * number of background threads (and stored into the {@link CachingDownloader}'s
 * cache. When ready to parse the new document, you can request the result by
 * calling {@link #waitFor(String)}, waiting for it if needed.
 */
public class PrefetchDownloader {

	private final CachingDownloader cd;
	private final Queue<PrefetchEntry> q = new LinkedList<PrefetchEntry>();
	private int busy = 0;
	private boolean finished = false;
	private final Map<String, InputStream> results;

	/**
	 * Create a new prefetch downloader.
	 * 
	 * @param cd
	 *            Caching downloader to use
	 * @param threadcount
	 *            Number of prefetch threads to start
	 * @param cacheResults
	 *            Whether to cache the results in the prefetcher. If not, they
	 *            are only written to the cache and can be retrieved from there
	 */
	public PrefetchDownloader(CachingDownloader cd, int threadcount, boolean cacheResults) {
		this.cd = cd;
		this.results = cacheResults ? new HashMap<String, InputStream>() : null;
		for (int i = 0; i < threadcount; i++) {
			new Thread(new Runnable() {
				@Override
				public void run() {
					runPrefetch();
				}
			}).start();
		}
	}

	/**
	 * Add a document to download
	 * 
	 * @param url
	 *            URL of the document
	 * @param cacheName
	 *            Name of the document in the cache
	 * 
	 * @see CachingDownloader#download(String, String)
	 */
	public synchronized void download(String url, String cacheName) {
		download(url, null, cacheName);
	}

	/**
	 * Add a document to download, including POST data.
	 * 
	 * @param url
	 *            URL of the document
	 * @param postData
	 *            postData to post
	 * @param cacheName
	 *            Name of the document in the cache
	 * 
	 * @see CachingDownloader#download(String, String, String)
	 */
	public synchronized void download(String url, String postData, String cacheName) {
		q.add(new PrefetchEntry(url, postData, cacheName));
		notifyAll();
	}

	/**
	 * Wait until all prefetching is finished. This is useful if the results do
	 * not have to be parsed, just downloaded, to make sure to wait until
	 * everything requested by now has been downloaded.
	 */
	public synchronized void waitPrefetch() throws InterruptedException {
		while (busy != 0 || !q.isEmpty())
			wait();
	}

	/**
	 * Wait until a given document has been prefetched.
	 * 
	 * @param cacheName
	 *            Name of the document in the cache
	 * @return The document content as a stream
	 */
	public synchronized InputStream waitFor(String cacheName) throws InterruptedException {
		if (results == null)
			throw new IllegalStateException("Result caching has been disabled");
		while (!results.containsKey(cacheName))
			wait();
		return results.remove(cacheName);
	}

	/**
	 * Wait until all prefetching is finished and shut down the prefetch
	 * threads.
	 */
	public synchronized void shutdown() throws InterruptedException {
		waitPrefetch();
		finished = true;
		notifyAll();
	}

	private void runPrefetch() {
		try {
			PrefetchEntry pe = null;
			while (true) {
				synchronized (this) {
					while (true) {
						if (finished)
							return;
						pe = q.poll();
						if (pe == null) {
							wait();
						} else {
							break;
						}
					}
					busy++;
				}
				InputStream in = cd.download(pe.url, pe.postData, pe.cacheName);
				synchronized (this) {
					busy--;
					if (results != null)
						results.put(pe.cacheName, in);
					notifyAll();
				}
			}
		} catch (Exception ex) {
			ex.printStackTrace();
		}
	}

	private static class PrefetchEntry {
		private final String url;
		private final String cacheName;
		private final String postData;

		public PrefetchEntry(String url, String postData, String cacheName) {
			this.url = url;
			this.postData = postData;
			this.cacheName = cacheName;
		}
	}
}