/*
 * Copyright (c) 2006 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util;

import java.util.regex.Pattern;

/**
 * Utility class for quoting regular expressions for matching literal text.
 * 
 * This class was very useful in Java 1.4, where {@link Pattern#quote(String)}
 * did not exist yet. Nowadays, only its {@link #quotePartially(String)}
 * function is useful.
 */
public class RegexQuoter {

	/**
	 * Quote an expression.
	 * 
	 * @param expression
	 *            any text
	 * @return a regular expression that matches exactly the input text
	 * @deprecated as of Java 1.5, use {@link Pattern#quote(String)} instead
	 */
	@Deprecated
	public static String quote(String expression) {
		return Pattern.quote(expression);
	}

	/**
	 * Quote an expression partially. Parts between &laquo; and &raquo; are not
	 * quoted, but left inside the expression.
	 */
	public static String quotePartially(String expression) {
		return quotePartially(expression, "", "");
	}

	/**
	 * Quote an expression partially. Parts between <i>startDelimiter</i> and
	 * <i>endDelimiter</i> are not quoted, but left inside the expression.
	 */
	public static String quotePartially(String expression, String startDelimiter, String endDelimiter) {
		StringBuilder sb = new StringBuilder();
		int pos;
		while ((pos = expression.indexOf(startDelimiter)) != -1) {
			sb.append(Pattern.quote(expression.substring(0, pos))).append("(");
			expression = expression.substring(pos + startDelimiter.length());
			pos = expression.indexOf(endDelimiter);
			if (pos == -1)
				throw new RuntimeException("Missing delimiter");
			if (pos == 0)
				sb.append(".*?");
			sb.append(expression.substring(0, pos)).append(")");
			expression = expression.substring(pos + endDelimiter.length());
		}
		return sb.append(Pattern.quote(expression)).toString();
	}
}
