/*
 * Copyright (c) 2006 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Reader;
import java.io.Writer;

/**
 * Utility class for forwarding data from an {@link InputStream} to an
 * {@link OutputStream} or from a {@link Reader} to a {@link Writer}.
 */
public class StreamForwarder extends Thread {

	/**
	 * Forward all content of an {@link InputStream} to an {@link OutputStream}.
	 * Both streams are closed afterwards.
	 * 
	 * @param in
	 *            the input stream
	 * @param out
	 *            the output stream
	 * @throws IOException
	 *             if one of the streams throws an {@link IOException}
	 */
	public static void forward(InputStream in, OutputStream out) throws IOException {
		try {
			final byte[] buf = new byte[4096];
			int length;
			while ((length = in.read(buf)) != -1) {
				if (out != null) {
					out.write(buf, 0, length);
					if (in.available() == 0) {
						out.flush();
					}
				}
			}
		} finally {
			in.close();
			out.close();
		}
	}

	/**
	 * Forward all content of a {@link Reader} to a {@link Writer}. Both are
	 * closed afterwards.
	 * 
	 * @param in
	 *            the reader
	 * @param out
	 *            the writer
	 * @throws IOException
	 *             if either the reader or the writer throws an
	 *             {@link IOException}
	 */
	public static void forward(Reader in, Writer out) throws IOException {
		try {
			final char[] buf = new char[4096];
			int length;
			while ((length = in.read(buf)) != -1) {
				if (out != null) {
					out.write(buf, 0, length);
					if (!in.ready()) {
						out.flush();
					}
				}
			}
		} finally {
			in.close();
			out.close();
		}
	}
}
