/*
 * Copyright (c) 2006 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util.csv;

import java.io.IOException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

/**
 * Abstract base class to read data from several CSV-like file formats.
 */
public abstract class AbstractCSVReader {

	/**
	 * Read a line from the CSV file, and pass its columns into a {@link Map}
	 * using the given keys/headers.
	 * 
	 * @param <T>
	 *            Type of the keys
	 * @param headers
	 *            Keys/Headers
	 * @param includeEmptyColumns
	 *            Whether to include empty column values in the map
	 */
	public final <T> Map<T, String> read(T[] headers, boolean includeEmptyColumns) throws IOException {
		Map<T, String> result = new HashMap<T, String>();
		String[] tempResult = read(headers.length);
		for (int i = 0; i < tempResult.length; i++) {
			if (tempResult[i].length() > 0 || includeEmptyColumns)
				result.put(headers[i], tempResult[i]);
		}
		return result;
	}

	/**
	 * Read a line from the CSV file and pad it to the given number of columns
	 * with empty strings.
	 * 
	 * @param columnCount
	 *            Desired number of columns
	 * @return The padded line
	 * @throws IOException
	 *             if there are too many columns in the original line.
	 */
	public final String[] read(int columnCount) throws IOException {
		String[] tempResult = read();
		if (tempResult.length == columnCount)
			return tempResult;
		if (tempResult.length > columnCount)
			throw new IOException("Too many columns: " + Arrays.asList(tempResult));
		String[] result = new String[columnCount];
		System.arraycopy(tempResult, 0, result, 0, tempResult.length);
		Arrays.fill(result, tempResult.length + 1, result.length, "");
		return result;
	}

	/**
	 * Write the complete content of this reader to the given
	 * {@link AbstractCSVWriter} and close both reader and writer.
	 * 
	 * @param writer
	 *            The writer to write to
	 */
	public void writeTo(AbstractCSVWriter writer) throws IOException {
		String[] record;
		while ((record = read()) != null)
			writer.write(record);
		close();
		writer.close();
	}

	/**
	 * Read a line from the CSV file.
	 */
	public abstract String[] read() throws IOException;

	/**
	 * Close this reader.
	 */
	public abstract void close() throws IOException;
}
