/*
 * Copyright (c) 2006 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util.csv;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.charset.Charset;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Properties;
import java.util.Set;

import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;

/**
 * Utility class to create CSV readers and writers based on an input file and a
 * format string that describes the format of the CSV file.
 */
public class CSVFactory {

	/**
	 * Create an {@link AbstractCSVReader} from a format string.
	 * 
	 * <h2>{@link CSVReader}</h2>
	 * <tt>separator=auto|comma|semicolon|tab|<i>char</i>[, flags=comments|quotedlines|multiline[+<i>...</i>]][, charset=<i>charset</i>]</tt>
	 * 
	 * <h2>{@link FixedWidthCSVReader}</h2>
	 * 
	 * <tt>fixed=<i>columnSpec1</i>[,<i>...</i>][, prefix=<i>prefix</i>][, separator=<i>separator</i>][, suffix=<i>suffix</i>][, flags=comments][, charset=<i>charset</i>]</tt>
	 * 
	 * <h2>{@link PropertiesCSVReader}</h2>
	 * 
	 * <tt>properties=<i>firstIndex</i>,<i>columnFormat1</i>[,<i>...</i>]</tt>
	 * 
	 * <h2>{@link XMLCSVReader}</h2>
	 * 
	 * <tt>xml=<i>rowTag</i>,<i>cellTag1</i>[,<i>...</i>]</tt>
	 * 
	 * 
	 * <h2>{@link XPathCSVReader}</h2>
	 * 
	 * <tt>xpath=<i>rowExpression</i>,<i>columnExpression1</i>[,<i>...</i>]</tt>
	 * 
	 * @param in
	 *            Input stream to read data from
	 * @param formatString
	 *            Format string
	 * @return The CSV reader
	 */
	public static AbstractCSVReader createReader(InputStream in, String formatString) throws IOException {
		return new CSVFactory(formatString, true).createReader(in);
	}

	/**
	 * Create an {@link AbstractCSVWriter} from a format string.
	 * 
	 * <h2>{@link CSVWriter}</h2>
	 * <tt>separator=comma|semicolon|tab|<i>char</i>[, charset=<i>charset</i>]</tt>
	 * 
	 * <h2>{@link FixedWidthCSVWriter}</h2>
	 * 
	 * <tt>fixed=<i>columnSpec1</i>[,<i>...</i>][, prefix=<i>prefix</i>][, separator=<i>separator</i>][, suffix=<i>suffix</i>][, flags=truncate][, charset=<i>charset</i>]</tt>
	 * 
	 * <h2>{@link PropertiesCSVWriter}</h2>
	 * 
	 * <tt>properties=<i>firstIndex</i>,<i>columnFormat1</i>[,<i>...</i>]</tt>
	 * 
	 * <h2>{@link XMLCSVWriter}</h2>
	 * 
	 * <tt>xml=<i>rowTag</i>,<i>cellTag1</i>[,<i>...</i>]</tt>
	 * 
	 * @param out
	 *            Output stream to write to
	 * @param formatString
	 *            Format string
	 * @return The CSV reader
	 */
	public static AbstractCSVWriter createWriter(OutputStream out, String formatString) throws IOException {
		return new CSVFactory(formatString, false).createWriter(out);
	}

	private final String charset;
	private final Format format;
	private final HashSet<String> flags;
	private final char separatorChar;
	private final FixedWidthColumn[] columns;
	private final String rowString;
	private final String[] columnStrings;

	private CSVFactory(String formatString, boolean forReader) {
		String charset = null;
		int pos = formatString.indexOf(", charset=");
		if (pos != -1) {
			charset = formatString.substring(pos + 10);
			formatString = formatString.substring(0, pos);
		}
		flags = new HashSet<String>();
		pos = formatString.indexOf(", flags=");
		if (pos != -1) {
			flags.addAll(Arrays.asList(formatString.substring(pos + 8).split("\\+")));
			formatString = formatString.substring(0, pos);
		}
		final String[] supportedFlags;
		final boolean charsetSupported;
		if (formatString.startsWith("separator=")) {
			String separatorName = formatString.substring(10);
			if (separatorName.equals("auto") && forReader) {
				separatorChar = 0;
			} else if (separatorName.equals("comma")) {
				separatorChar = ',';
			} else if (separatorName.equals("semicolon")) {
				separatorChar = ';';
			} else if (separatorName.equals("tab")) {
				separatorChar = '\t';
			} else if (separatorName.length() == 1) {
				separatorChar = separatorName.charAt(0);
			} else {
				throw new IllegalArgumentException("Unsupported separator: " + separatorName);
			}
			charsetSupported = true;
			rowString = null;
			columns = null;
			columnStrings = null;
			supportedFlags = forReader ? new String[] { "comments", "quotedlines", "multiline" } : new String[0];
			format = Format.CSV;
		} else if (formatString.startsWith("fixed=")) {
			String prefix = "", separator = "", suffix = "";
			pos = formatString.indexOf(", suffix=");
			if (pos != -1) {
				suffix = formatString.substring(pos + 9);
				formatString = formatString.substring(0, pos);
			}
			pos = formatString.indexOf(", separator=");
			if (pos != -1) {
				separator = formatString.substring(pos + 12);
				formatString = formatString.substring(0, pos);
			}
			pos = formatString.indexOf(", prefix=");
			if (pos != -1) {
				prefix = formatString.substring(pos + 9);
				formatString = formatString.substring(0, pos);
			}
			columnStrings = new String[] { prefix, separator, suffix };
			String[] colSpecs = formatString.substring(6).split(",");
			columns = new FixedWidthColumn[colSpecs.length];
			for (int i = 0; i < colSpecs.length; i++) {
				columns[i] = FixedWidthColumn.parse(colSpecs[i]);
			}
			charsetSupported = true;
			rowString = null;
			separatorChar = 0;
			supportedFlags = forReader ? new String[] { "comments" } : new String[] { "truncate" };
			format = Format.FIXED;
		} else {
			String[] parts;
			if (formatString.startsWith("properties=")) {
				parts = formatString.substring(11).split(",", 2);
				format = Format.PROPERTIES;
			} else if (formatString.startsWith("xml=")) {
				parts = formatString.substring(4).split(",", 2);
				format = Format.XML;
			} else if (formatString.startsWith("xpath=") && forReader) {
				parts = formatString.substring(6).split(",", 2);
				format = Format.XPATH;
			} else {
				throw new IllegalArgumentException("Unparsable format string: " + formatString);
			}
			rowString = parts[0];
			columnStrings = parts[1].split(",");
			charsetSupported = false;
			supportedFlags = new String[0];
			columns = null;
			separatorChar = 0;
		}

		if (charsetSupported) {
			if (charset == null)
				charset = "UTF-BOM." + Charset.defaultCharset().name();
		} else if (charset != null) {
			throw new IllegalArgumentException("Charset not supported for this format");
		}
		this.charset = charset;
		Set<String> supportedFlagsSet = new HashSet<String>();
		supportedFlagsSet.addAll(Arrays.asList(supportedFlags));
		for (String flag : flags) {
			if (!supportedFlagsSet.contains(flag))
				throw new IllegalArgumentException("Unsupported flag: " + flag);
		}
	}

	private AbstractCSVReader createReader(InputStream in) throws IOException {
		switch (format) {
		case CSV: {
			CSVReader r = new CSVReader(in, charset);
			r.setSeparator(separatorChar);
			r.setStripComments(flags.contains("comments"));
			r.setSupportFullyQuotedLines(flags.contains("quotedlines"));
			r.setSupportMultiLineCells(flags.contains("multiline"));
			return r;
		}
		case FIXED: {
			final String prefix = columnStrings[0], separator = columnStrings[1], suffix = columnStrings[2];
			FixedWidthCSVReader r = new FixedWidthCSVReader(in, charset, prefix, separator, suffix, columns);
			r.setStripComments(flags.contains("comments"));
			return r;
		}
		case PROPERTIES: {
			Properties props = new Properties();
			props.load(in);
			in.close();
			return new PropertiesCSVReader(props, Integer.parseInt(rowString), columnStrings);
		}
		case XML: {
			try {
				Document doc = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(in);
				return new XMLCSVReader(doc, rowString, columnStrings);
			} catch (Exception ex) {
				throw new IOException("Exception while parsing XML file", ex);
			}
		}
		case XPATH: {
			try {
				Document doc = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(in);
				return new XPathCSVReader(doc, rowString, columnStrings);
			} catch (Exception ex) {
				throw new IOException("Exception while parsing XML file", ex);
			}
		}
		default:
			throw new IOException("Unsupported CSV format");
		}
	}

	private AbstractCSVWriter createWriter(OutputStream out) throws IOException {
		switch (format) {
		case CSV: {
			return new CSVWriter(out, charset, separatorChar);
		}
		case FIXED: {
			final String prefix = columnStrings[0], separator = columnStrings[1], suffix = columnStrings[2];
			FixedWidthCSVWriter w = new FixedWidthCSVWriter(out, charset, prefix, separator, suffix, columns);
			w.setTruncateValues(flags.contains("truncate"));
			return w;
		}
		case PROPERTIES: {
			return new PropertiesCSVWriter(out, Integer.parseInt(rowString), columnStrings);
		}
		case XML: {
			try {
				return new XMLCSVWriter(out, "csv", rowString, columnStrings);
			} catch (Exception ex) {
				throw new IOException("Exception while creating XML file", ex);
			}
		}
		default:
			throw new IOException("Unsupported CSV format");
		}
	}

	private enum Format {
		CSV, FIXED, PROPERTIES, XML, XPATH
	}
}
