/*
 * Copyright (c) 2006 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util.csv;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.List;

/**
 * Class to read data from a CSV (Comma Separated Value) file, supporting
 * different separator characters and quoted values.
 */
public class CSVReader extends AbstractCSVReader {

	private final BufferedReader br;
	private char separator = 0;
	private boolean stripComments = false;
	private boolean supportFullyQuotedLines = false;
	private boolean supportMultiLineCells = false;

	/**
	 * Create a new {@link CSVReader} that reads from the given buffered reader.
	 */
	public CSVReader(BufferedReader br) {
		this.br = br;
	}

	/**
	 * Create a new {@link CSVReader} that reads from the given reader.
	 */
	public CSVReader(Reader r) {
		this(new BufferedReader(r));
	}

	/**
	 * Create a new {@link CSVReader} that reads from the given stream.
	 */
	public CSVReader(InputStream in, String charsetName) throws UnsupportedEncodingException {
		this(new InputStreamReader(in, charsetName));
	}

	/**
	 * Return the separator character (usually comma, tab or semicolon), or 0 to
	 * auto-detect.
	 */
	public char getSeparator() {
		return separator;
	}

	/**
	 * Set the separator character (usually comma, tab or semicolon), or 0 to
	 * auto-detect.
	 */
	public void setSeparator(char separator) {
		this.separator = separator;
	}

	/**
	 * Set whether comments (lines starting with # or empty lines) should be
	 * stripped.
	 */
	public void setStripComments(boolean stripComments) {
		this.stripComments = stripComments;
	}

	/**
	 * Some broken CSV writers quote the whole line if no field of the line
	 * needs to be quoted. In that case, a standard-compliant CSV parser will
	 * read everything into one field. To support these files, this option can
	 * be enabled. In that case, if there is only a quote character at the
	 * beginning and the end of the line, it is stripped.
	 */
	public void setSupportFullyQuotedLines(boolean supportFullyQuotedLines) {
		this.supportFullyQuotedLines = supportFullyQuotedLines;
	}

	/**
	 * Enable support for multi-line values in cells. In that case, if a
	 * quotation mark is not closed within the same line, more lines are read to
	 * find the closing quotation mark. Lines will be combined with newline
	 * characters <tt>\n</tt> regardless of the original line delimiter
	 * character.
	 * 
	 * If disabled, quotation marks that are not closed within the same line are
	 * treated as a parse error.
	 */
	public void setSupportMultiLineCells(boolean supportMutiLineCells) {
		this.supportMultiLineCells = supportMutiLineCells;
	}

	@Override
	public void close() throws IOException {
		br.close();
	}

	@Override
	public String[] read() throws IOException {
		String line = readNextLine();
		if (line == null)
			return null;
		if (separator == 0) {
			if (line.indexOf('\t') != -1)
				separator = '\t';
			else if (line.indexOf(';') != -1)
				separator = ';';
			else if (line.indexOf(',') != -1)
				separator = ',';
			else
				separator = ';';
		}
		final String origLine = line;
		List<String> elems = new ArrayList<String>();
		while (line != null) {
			if (line.startsWith("\"")) {
				String part = "";
				while (true) {
					int pos = line.indexOf("\"", 1);
					while (supportMultiLineCells && pos == -1) {
						String nextLine = readNextLine();
						if (nextLine == null)
							break;
						line = line + '\n' + nextLine;
						pos = line.indexOf("\"", 1);
					}
					if (pos == -1)
						throw new IOException("CSV line cannot be parsed: " + origLine);
					part += line.substring(1, pos);
					line = line.substring(pos + 1);
					if (line.length() == 0) {
						line = null;
						break;
					} else if (line.charAt(0) == separator) {
						line = line.substring(1);
						break;
					} else if (line.charAt(0) != '"') {
						throw new IOException("CSV line cannot be parsed: " + origLine + "\nat: " + line);
					}
					part += "\"";
				}
				elems.add(part);
			} else {
				int pos = line.indexOf(separator);
				if (pos == -1) {
					elems.add(line);
					line = null;
				} else {
					elems.add(line.substring(0, pos));
					line = line.substring(pos + 1);
				}
			}
		}
		return elems.toArray(new String[elems.size()]);
	}

	private String readNextLine() throws IOException {
		String line = br.readLine();
		if (stripComments) {
			while (line != null && (line.startsWith("#") || line.length() == 0))
				line = br.readLine();
		}
		if (line != null && supportFullyQuotedLines && line.length() > 2 && line.startsWith("\"") && line.endsWith("\"") && line.substring(1, line.length() - 1).indexOf('"') == -1) {
			line = line.substring(1, line.length() - 1);
		}
		return line;
	}
}
