/*
 * Copyright (c) 2010 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util.csv;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.UnsupportedEncodingException;

/**
 * Class to read data from a fixed width CSV (Comma Separated Value) file, where
 * every column has a fixed width and may optionally additionally delimited by
 * separator characters.
 */
public class FixedWidthCSVReader extends AbstractCSVReader {

	private final BufferedReader br;
	private final String prefix;
	private final String separator;
	private final String suffix;
	private final FixedWidthColumn[] columns;
	private final int lineLength;
	private boolean stripComments = false;

	/**
	 * Create a new {@link FixedWidthCSVReader} that reads from the given
	 * buffered reader using left-aligned space-padded columns only and no
	 * separators.
	 * 
	 * @param br
	 *            The buffered reader to read from
	 * @param widths
	 *            The width of each column
	 */
	public FixedWidthCSVReader(BufferedReader br, int... widths) {
		this(br, buildColumns(widths));
	}

	static FixedWidthColumn[] buildColumns(int[] widths) {
		FixedWidthColumn[] result = new FixedWidthColumn[widths.length];
		for (int i = 0; i < widths.length; i++) {
			result[i] = new FixedWidthColumn(widths[i]);
		}
		return result;
	}

	/**
	 * Create a new {@link FixedWidthCSVReader} that reads from the given
	 * buffered reader using columns parsed from the given column specs and no
	 * separators.
	 * 
	 * @param br
	 *            The buffered reader to read from
	 * @param columnSpecs
	 *            The format specification of each column
	 */
	public FixedWidthCSVReader(BufferedReader br, String... columnSpecs) throws IOException {
		this(br, buildColumns(columnSpecs));
	}

	static FixedWidthColumn[] buildColumns(String[] columnSpecs) {
		FixedWidthColumn[] result = new FixedWidthColumn[columnSpecs.length];
		for (int i = 0; i < columnSpecs.length; i++) {
			result[i] = FixedWidthColumn.parse(columnSpecs[i]);
		}
		return result;
	}

	/**
	 * Create a new {@link FixedWidthCSVReader} that reads from the given
	 * buffered reader using the given columns and no separators.
	 * 
	 * @param br
	 *            The buffered reader to read from
	 * @param columns
	 *            The format of each column
	 */
	public FixedWidthCSVReader(BufferedReader br, FixedWidthColumn... columns) {
		this(br, "", "", "", columns);
	}

	/**
	 * Create a new {@link FixedWidthCSVReader} that reads from the given
	 * buffered reader using the given columns and separators.
	 * 
	 * @param br
	 *            The buffered reader to read from
	 * @param prefix
	 *            The prefix to add to each line
	 * @param separator
	 *            The separator to add between each field
	 * @param suffix
	 *            The suffix to add to each line
	 * @param columns
	 *            The format of each column
	 */
	public FixedWidthCSVReader(BufferedReader br, String prefix, String separator, String suffix, FixedWidthColumn... columns) {
		this.br = br;
		this.prefix = prefix;
		this.separator = separator;
		this.suffix = suffix;
		this.columns = columns;
		int lineLength = prefix.length() + suffix.length() + separator.length() * (columns.length - 1);
		for (int i = 0; i < columns.length; i++) {
			lineLength += columns[i].getWidth();
		}
		this.lineLength = lineLength;
	}

	/**
	 * Create a new {@link FixedWidthCSVReader} that reads from the given reader
	 * using left-aligned space-padded columns only and no separators.
	 * 
	 * @param r
	 *            The reader to read from
	 * @param widths
	 *            The width of each column
	 */
	public FixedWidthCSVReader(Reader r, int... widths) {
		this(new BufferedReader(r), widths);
	}

	/**
	 * Create a new {@link FixedWidthCSVReader} that reads from the given reader
	 * using columns parsed from the given column specs and no separators.
	 * 
	 * @param r
	 *            The reader to read from
	 * @param columnSpecs
	 *            The format specification of each column
	 */
	public FixedWidthCSVReader(Reader r, String... columnSpecs) throws IOException {
		this(new BufferedReader(r), columnSpecs);
	}

	/**
	 * Create a new {@link FixedWidthCSVReader} that reads from the given reader
	 * using the given columns and no separators.
	 * 
	 * @param r
	 *            The reader to read from
	 * @param columns
	 *            The format of each column
	 */
	public FixedWidthCSVReader(Reader r, FixedWidthColumn... columns) {
		this(new BufferedReader(r), columns);
	}

	/**
	 * Create a new {@link FixedWidthCSVReader} that reads from the given reader
	 * using the given columns and separators.
	 * 
	 * @param r
	 *            The reader to read from
	 * @param prefix
	 *            The prefix to add to each line
	 * @param separator
	 *            The separator to add between each field
	 * @param suffix
	 *            The suffix to add to each line
	 * @param columns
	 *            The format of each column
	 */
	public FixedWidthCSVReader(Reader r, String prefix, String separator, String suffix, FixedWidthColumn... columns) {
		this(new BufferedReader(r), prefix, separator, suffix, columns);
	}

	/**
	 * Create a new {@link FixedWidthCSVReader} that reads from the given stream
	 * using left-aligned space-padded columns only and no separators.
	 * 
	 * @param in
	 *            The input stream to read from
	 * @param charsetName
	 *            Name of the charset to be used
	 * @param widths
	 *            The width of each column
	 */
	public FixedWidthCSVReader(InputStream in, String charsetName, int... widths) throws UnsupportedEncodingException {
		this(new InputStreamReader(in, charsetName), widths);
	}

	/**
	 * Create a new {@link FixedWidthCSVReader} that reads from the given stream
	 * using columns parsed from the given column specs and no separators.
	 * 
	 * @param in
	 *            The input stream to read from
	 * @param charsetName
	 *            Name of the charset to be used
	 * @param columnSpecs
	 *            The format specification of each column
	 */
	public FixedWidthCSVReader(InputStream in, String charsetName, String... columnSpecs) throws IOException {
		this(new InputStreamReader(in, charsetName), columnSpecs);
	}

	/**
	 * Create a new {@link FixedWidthCSVReader} that reads from the given stream
	 * using the given columns and no separators.
	 * 
	 * @param in
	 *            The input stream to read from
	 * @param charsetName
	 *            Name of the charset to be used
	 * @param columns
	 *            The format of each column
	 */
	public FixedWidthCSVReader(InputStream in, String charsetName, FixedWidthColumn... columns) throws UnsupportedEncodingException {
		this(new InputStreamReader(in, charsetName), columns);
	}

	/**
	 * Create a new {@link FixedWidthCSVReader} that reads from the given stream
	 * using the given columns and separators.
	 * 
	 * @param in
	 *            The input stream to read from
	 * @param charsetName
	 *            Name of the charset to be used
	 * @param prefix
	 *            The prefix to add to each line
	 * @param separator
	 *            The separator to add between each field
	 * @param suffix
	 *            The suffix to add to each line
	 * @param columns
	 *            The format of each column
	 */
	public FixedWidthCSVReader(InputStream in, String charsetName, String prefix, String separator, String suffix, FixedWidthColumn... columns) throws UnsupportedEncodingException {
		this(new InputStreamReader(in, charsetName), prefix, separator, suffix, columns);
	}

	@Override
	public void close() throws IOException {
		br.close();
	}

	/**
	 * Set whether comments (lines starting with # or empty lines) should be
	 * stripped.
	 */
	public void setStripComments(boolean stripComments) {
		this.stripComments = stripComments;
	}

	/**
	 * Read a line from the fixed width CSV file.
	 */
	@Override
	public String[] read() throws IOException {
		String line = br.readLine();
		if (stripComments) {
			while (line != null && (line.startsWith("#") || line.length() == 0))
				line = br.readLine();
		}
		if (line == null)
			return null;
		if (line.length() != lineLength)
			throw new IOException("Invalid line length: " + line);
		if (!line.startsWith(prefix))
			throw new IOException("Invalid prefix: " + line);
		if (!line.endsWith(suffix))
			throw new IOException("Invalid suffix: " + line);
		final String origLine = line;
		line = separator + line.substring(prefix.length(), line.length() - suffix.length());
		String[] elems = new String[columns.length];
		for (int i = 0; i < elems.length; i++) {
			if (!line.startsWith(separator))
				throw new IOException("Separator for column " + i + " missing: " + line);
			line = line.substring(separator.length());
			elems[i] = columns[i].parseValue(line.substring(0, columns[i].getWidth()));
			line = line.substring(columns[i].getWidth());
		}
		if (line.length() != 0)
			throw new RuntimeException("Lengths did not add up: " + origLine);
		return elems;
	}
}