/*
 * Copyright (c) 2010 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util.csv;

import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.io.Writer;

/**
 * Class to create fixed width CSV (Comma Separated Value) files, where every
 * column has a fixed width and may optionally additionally delimited by
 * separator characters.
 */
public class FixedWidthCSVWriter extends AbstractCSVWriter {

	private final Writer w;
	private final String prefix;
	private final String separator;
	private final String suffix;
	private final FixedWidthColumn[] columns;
	private boolean truncateValues;

	/**
	 * Create a new CSV writer that writes to the given {@link Writer} using
	 * left-aligned space-padded columns only and no separators.
	 * 
	 * @param w
	 *            the writer to write to.
	 * @param widths
	 *            The width of each column
	 */
	public FixedWidthCSVWriter(Writer w, int... widths) {
		this(w, FixedWidthCSVReader.buildColumns(widths));
	}

	/**
	 * Create a new CSV writer that writes to the given {@link Writer} using
	 * columns parsed from the given column specs and no separators.
	 * 
	 * @param w
	 *            the writer to write to.
	 * @param columnSpecs
	 *            The format specification of each column
	 */
	public FixedWidthCSVWriter(Writer w, String... columnSpecs) {
		this(w, FixedWidthCSVReader.buildColumns(columnSpecs));
	}

	/**
	 * Create a new CSV writer that writes to the given {@link Writer} using the
	 * given columns and no separators.
	 * 
	 * @param w
	 *            the writer to write to.
	 * @param columns
	 *            The format of each column
	 */
	public FixedWidthCSVWriter(Writer w, FixedWidthColumn... columns) {
		this(w, "", "", "", columns);
	}

	/**
	 * Create a new CSV writer that writes to the given {@link Writer} using the
	 * given columns and separators.
	 * 
	 * @param w
	 *            the writer to write to.
	 * @param prefix
	 *            The prefix to add to each line
	 * @param separator
	 *            The separator to add between each field
	 * @param suffix
	 *            The suffix to add to each line
	 * @param columns
	 *            The format of each column
	 */
	public FixedWidthCSVWriter(Writer w, String prefix, String separator, String suffix, FixedWidthColumn... columns) {
		this.w = w;
		this.prefix = prefix;
		this.separator = separator;
		this.suffix = suffix;
		this.columns = columns;
	}

	/**
	 * Create a new CSV writer that writes to the given stream in the given
	 * charset using left-aligned space-padded columns only and no separators.
	 * 
	 * @param out
	 *            Output stream
	 * @param charsetName
	 *            charset
	 * @param widths
	 *            The width of each column
	 */
	public FixedWidthCSVWriter(OutputStream out, String charsetName, int... widths) throws UnsupportedEncodingException {
		this(new OutputStreamWriter(out, charsetName), widths);
	}

	/**
	 * Create a new CSV writer that writes to the given stream in the given
	 * charset using columns parsed from the given column specs and no
	 * separators.
	 * 
	 * @param out
	 *            Output stream
	 * @param charsetName
	 *            charset
	 * @param columnSpecs
	 *            The format specification of each column
	 */
	public FixedWidthCSVWriter(OutputStream out, String charsetName, String... columnSpecs) throws UnsupportedEncodingException {
		this(new OutputStreamWriter(out, charsetName), columnSpecs);
	}

	/**
	 * Create a new CSV writer that writes to the given stream in the given
	 * charset using the given columns and no separators.
	 * 
	 * @param out
	 *            Output stream
	 * @param charsetName
	 *            charset
	 * @param columns
	 *            The format of each column
	 */
	public FixedWidthCSVWriter(OutputStream out, String charsetName, FixedWidthColumn... columns) throws UnsupportedEncodingException {
		this(new OutputStreamWriter(out, charsetName), columns);
	}

	/**
	 * Create a new CSV writer that writes to the given stream in the given
	 * charset using the given columns and separators.
	 * 
	 * @param out
	 *            Output stream
	 * @param charsetName
	 *            charset
	 * @param prefix
	 *            The prefix to add to each line
	 * @param separator
	 *            The separator to add between each field
	 * @param suffix
	 *            The suffix to add to each line
	 * @param columns
	 *            The format of each column
	 */
	public FixedWidthCSVWriter(OutputStream out, String charsetName, String prefix, String separator, String suffix, FixedWidthColumn... columns) throws UnsupportedEncodingException {
		this(new OutputStreamWriter(out, charsetName), prefix, separator, suffix, columns);
	}

	@Override
	public void flush() throws IOException {
		w.flush();
	}

	@Override
	public void close() throws IOException {
		w.close();
	}

	/**
	 * Set whether overlong values should be truncated automatically. If
	 * disabled, an exception will be thrown.
	 */
	public void setTruncateValues(boolean truncateValues) {
		this.truncateValues = truncateValues;
	}

	@Override
	public void write(String... record) throws IOException {
		if (record.length > columns.length)
			throw new IOException("Too many columns");
		for (int i = 0; i < columns.length; i++) {
			if (i == 0)
				w.write(prefix);
			else
				w.write(separator);
			w.write(columns[i].formatValue(i < record.length ? record[i] : "", truncateValues, false));
		}
		w.write(suffix + "\r\n");
	}
}
