/*
 * Copyright (c) 2010 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util.csv;

import java.io.IOException;
import java.util.Arrays;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * A column of a fixed-width CSV file.
 */
public class FixedWidthColumn {

	private final int width;
	private final Alignment alignment;
	private final char padLeft;
	private final char padRight;
	private final String[] exceptionStrings;

	/**
	 * Create a left-aligned fixed-width column of the given width.
	 * 
	 * Spaces are used to pad the value on both sides. Values that contain tabs
	 * or two consecutive spaces are considered invalid.
	 * 
	 * @param width
	 *            The width
	 */
	public FixedWidthColumn(int width) {
		this(width, Alignment.LEFT);
	}

	/**
	 * Create a fixed-width column of the given width and alignment.
	 * 
	 * Spaces are used to pad the value on both sides. Values that contain tabs
	 * or two consecutive spaces are considered invalid.
	 * 
	 * @param width
	 *            The width
	 * @param alignment
	 *            The alignment
	 */
	public FixedWidthColumn(int width, Alignment alignment) {
		this(width, alignment, ' ');
	}

	/**
	 * Create a fixed-width column of the given width and alignment, which uses
	 * the given character to pad the value on both sides if needed.
	 * 
	 * Values that contain tabs or two consecutive spaces are considered
	 * invalid.
	 * 
	 * @param width
	 *            The width
	 * @param alignment
	 *            The alignment
	 * @param pad
	 *            The padding character
	 */
	public FixedWidthColumn(int width, Alignment alignment, char pad) {
		this(width, alignment, pad, pad);
	}

	/**
	 * Create a fixed-width column of the given width and alignment, which uses
	 * two different characters to pad the value on left or right side if
	 * needed.
	 * 
	 * Values that contain tabs or two consecutive spaces are considered
	 * invalid.
	 * 
	 * @param width
	 *            The width
	 * @param alignment
	 *            The alignment
	 * @param padLeft
	 *            The left padding character
	 * @param padRight
	 *            The right padding character
	 */
	public FixedWidthColumn(int width, Alignment alignment, char padLeft, char padRight) {
		this(width, alignment, padLeft, padRight, "  ", "\t");
	}

	/**
	 * Create a fixed-width column of the given width and alignment, which uses
	 * two different characters to pad the value on left or right side if needed
	 * and the given exception strings.
	 * 
	 * To use an empty exception string array, the empty array has to be given
	 * explicitly to not clash with the
	 * {@link #FixedWidthColumn(int, Alignment, char, char)} constructor.
	 * 
	 * Values that contain one of the given exception strings are considered
	 * invalid.
	 * 
	 * @param width
	 *            The width
	 * @param alignment
	 *            The alignment
	 * @param padLeft
	 *            The left padding character
	 * @param padRight
	 *            The right padding character
	 * @param exceptionStrings
	 *            The exception strings (may not contain null or empty strings!)
	 */
	public FixedWidthColumn(int width, Alignment alignment, char padLeft, char padRight, String... exceptionStrings) {
		this.width = width;
		this.alignment = alignment;
		this.padLeft = padLeft;
		this.padRight = padRight;
		this.exceptionStrings = exceptionStrings;
		for (int i = 0; i < exceptionStrings.length; i++) {
			if (exceptionStrings[i] == null || exceptionStrings[i].length() == 0)
				throw new IllegalArgumentException("Illegal exception string");
		}
	}

	/**
	 * Parse a column from a column specification string.
	 * 
	 * <p>
	 * The specification starts with the column width as a decimal number,
	 * followed by an optional LRTB for the alignment, followed by an optional
	 * slash followed by one or two padding characters. If two padding
	 * characters are given, another slash delimits exception strings (each one
	 * delimited by yet another slash). An empty exception list (but with the
	 * leading slash) is treated as no exceptions.
	 * 
	 * <p>
	 * A missing parameter is treated as the default as used by the
	 * {@link #FixedWidthColumn(int)} constructor.
	 * 
	 * @param colSpec
	 *            column specification
	 * @return the column
	 */
	public static FixedWidthColumn parse(String colSpec) {
		Pattern colRegex = Pattern.compile("([0-9]+)([LRCF]?)(?:(/..?)(/.*)?)?");
		Matcher m = colRegex.matcher(colSpec);
		if (!m.matches())
			throw new IllegalArgumentException("Invalid column spec: " + colSpec);
		int width = Integer.parseInt(m.group(1));
		int alignmentIndex = "LRCF".indexOf(m.group(2));
		if (alignmentIndex == -1)
			alignmentIndex = 0;
		Alignment alignment = new Alignment[] { Alignment.LEFT, Alignment.RIGHT, Alignment.CENTER, Alignment.FILL }[alignmentIndex];
		char padLeft = ' ', padRight = ' ';
		String padGroup = m.group(3);
		if (padGroup != null) {
			padLeft = padGroup.charAt(1);
			padRight = padGroup.charAt(padGroup.length() - 1);
		}
		String[] exceptionStrings = new String[] { "  ", "\t" };
		String exceptionGroup = m.group(4);
		if (exceptionGroup != null) {
			exceptionStrings = exceptionGroup.substring(1).split("/");
			if (exceptionGroup.length() == 1)
				exceptionStrings = new String[0];
		}
		return new FixedWidthColumn(width, alignment, padLeft, padRight, exceptionStrings);
	}

	/**
	 * Parse a value from a fixed-width CSV file, i. e. remove padding and check
	 * for exception strings.
	 * 
	 * @param value
	 *            Value to parse
	 * @return Parsed value
	 */
	public String parseValue(String value) {
		if (alignment.isPadLeft()) {
			int first = 0;
			for (int i = 0; i < value.length(); i++) {
				if (value.charAt(i) != padLeft)
					break;
				first = i + 1;
			}
			value = value.substring(first);
		}
		if (alignment.isPadRight()) {
			int last = value.length();
			for (int i = value.length() - 1; i >= 0; i--) {
				if (value.charAt(i) != padRight)
					break;
				last = i;
			}
			value = value.substring(0, last);
		}
		for (String exception : exceptionStrings) {
			if (value.contains(exception))
				throw new IllegalArgumentException("Invalid column value: " + value);
		}
		return value;
	}

	/**
	 * Format a value to store it into a fixed-width CSV file. The value is
	 * padded as needed.
	 * 
	 * Overlong values can be truncated (on the right unless the column is
	 * right-aligned), left as is or cause an exception, based on your needs.
	 * 
	 * @param value
	 *            Value to format
	 * @param truncate
	 *            Whether to silently truncate overlong values
	 * @param allowOverflow
	 *            Whether to allow overflow for long values
	 * @return Formatted value
	 */
	public String formatValue(String value, boolean truncate, boolean allowOverflow) throws IOException {
		if (value.length() > width && truncate && alignment == Alignment.RIGHT)
			value = value.substring(value.length() - width);
		if (value.length() > width && truncate)
			value = value.substring(0, width);
		if (value.length() == width)
			return value;
		if (value.length() > width && !allowOverflow)
			throw new IOException("Value too long: " + value);
		StringBuilder sb = new StringBuilder(value);
		while (sb.length() < width && alignment != Alignment.FILL) {
			if (alignment.isPadRight())
				sb.append(padRight);
			if (sb.length() == width)
				break;
			if (alignment.isPadLeft())
				sb.insert(0, padLeft);
		}
		if (sb.length() < width && !allowOverflow)
			throw new IOException("Value too short: " + value);
		return sb.toString();
	}

	/**
	 * Get the width of this column.
	 */
	public int getWidth() {
		return width;
	}

	@Override
	public String toString() {
		return "FixedWidthColumn [width=" + width + ", alignment=" + alignment + ", padLeft=" + padLeft + ", padRight=" + padRight + ", exceptionStrings=" + exceptionStrings.length + " " + Arrays.toString(exceptionStrings) + "]";
	}

	/**
	 * Enumeration of possible column alignments.
	 */
	public static enum Alignment {

		/**
		 * Left alignment, pad on the right.
		 */
		LEFT(false, true),

		/**
		 * Right alignment, pad on the left.
		 */
		RIGHT(true, false),

		/**
		 * Center alignment, pad on both sides. When an odd number of padding is
		 * needed, pad one more on the right side.
		 */
		CENTER(true, true),

		/**
		 * Filled alignment, do not pad. Mostly used for reading files so that
		 * no padding is removed.
		 */
		FILL(false, false);

		private final boolean padLeft;
		private final boolean padRight;

		private Alignment(boolean padLeft, boolean padRight) {
			this.padLeft = padLeft;
			this.padRight = padRight;
		}

		boolean isPadLeft() {
			return padLeft;
		}

		boolean isPadRight() {
			return padRight;
		}
	}
}
