/*
 * Copyright (c) 2010 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util.csv;

import java.io.IOException;
import java.text.Format;
import java.text.MessageFormat;
import java.util.Properties;

/**
 * Class to read data from a {@link Properties} file as if it was a CSV (Comma
 * Separated Value) file.
 */
public class PropertiesCSVReader extends AbstractCSVReader {

	private final Properties properties;
	private final Format[] columnFormats;
	private int nextIndex = 0;

	/**
	 * Create a new {@link PropertiesCSVReader} that reads from the given
	 * {@link Properties} file. Each columnFormat is evaluated for each row
	 * number to return the respective property name for this column.
	 * 
	 * @param properties
	 *            Property file to write to
	 * @param firstIndex
	 *            index of the first row
	 * @param columnFormats
	 *            column formats
	 */
	public PropertiesCSVReader(Properties properties, int firstIndex, String... columnFormats) {
		this(properties, firstIndex, buildFormats(columnFormats));
	}

	/**
	 * Create a new {@link PropertiesCSVReader} that reads from the given
	 * {@link Properties} file. Each columnFormat is evaluated for each row
	 * number to return the respective property name for this column.
	 * 
	 * @param properties
	 *            Property file to write to
	 * @param firstIndex
	 *            index of the first row
	 * @param columnFormats
	 *            column formats
	 */
	public PropertiesCSVReader(Properties properties, int firstIndex, Format... columnFormats) {
		this.properties = properties;
		nextIndex = firstIndex;
		this.columnFormats = columnFormats;
	}

	static Format[] buildFormats(String... formats) {
		Format[] result = new Format[formats.length];
		for (int i = 0; i < result.length; i++) {
			result[i] = new MessageFormat(formats[i]);
		}
		return result;
	}

	/**
	 * Close this reader. This is a no-op.
	 */
	@Override
	public void close() throws IOException {
	}

	/**
	 * Read a row from this XML file.
	 */
	@Override
	public String[] read() throws IOException {
		String[] record = new String[columnFormats.length];
		boolean found = false;
		for (int i = 0; i < record.length; i++) {
			record[i] = properties.getProperty(columnFormats[i].format(new Object[] { nextIndex }));
			if (record[i] == null)
				record[i] = "";
			else
				found = true;
		}
		if (!found)
			return null;
		nextIndex++;
		return record;
	}
}