/*
 * Copyright (c) 2010 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util.csv;

import java.io.IOException;
import java.io.OutputStream;
import java.text.Format;
import java.util.Properties;

/**
 * Class to create a {@link Properties} file as if it was a CSV (Comma Separated
 * Value) file.
 */
public class PropertiesCSVWriter extends AbstractCSVWriter {

	private final OutputStream out;
	private final Properties properties;
	private final Format[] columnFormats;
	private int nextIndex = 0;

	/**
	 * Create a new {@link PropertiesCSVWriter} that writes to the given
	 * {@link Properties} file. Each columnFormat is evaluated for each row
	 * number to return the respective property name for this column.
	 * 
	 * @param properties
	 *            Property file to write to
	 * @param firstIndex
	 *            index of the first row
	 * @param columnFormats
	 *            column formats
	 */
	public PropertiesCSVWriter(Properties properties, int firstIndex, String... columnFormats) {
		this(properties, firstIndex, PropertiesCSVReader.buildFormats(columnFormats));
	}

	/**
	 * Create a new {@link PropertiesCSVWriter} that writes to the given
	 * {@link Properties} file. Each columnFormat is evaluated for each row
	 * number to return the respective property name for this column.
	 * 
	 * @param properties
	 *            Property file to write to
	 * @param firstIndex
	 *            index of the first row
	 * @param columnFormats
	 *            column formats
	 */
	public PropertiesCSVWriter(Properties properties, int firstIndex, Format... columnFormats) {
		this.out = null;
		this.properties = properties;
		this.nextIndex = firstIndex;
		this.columnFormats = columnFormats;
	}

	/**
	 * Create a new {@link PropertiesCSVWriter} that writes to the given output
	 * stream. Each columnFormat is evaluated for each row number to return the
	 * respective property name for this column.
	 * 
	 * @param out
	 *            Output stream to write to
	 * @param firstIndex
	 *            index of the first row
	 * @param columnFormats
	 *            column formats
	 */
	public PropertiesCSVWriter(OutputStream out, int firstIndex, String... columnFormats) {
		this(out, firstIndex, PropertiesCSVReader.buildFormats(columnFormats));
	}

	/**
	 * Create a new {@link PropertiesCSVWriter} that writes to the given output
	 * stream. Each columnFormat is evaluated for each row number to return the
	 * respective property name for this column.
	 * 
	 * @param out
	 *            Output stream to write to
	 * @param firstIndex
	 *            index of the first row
	 * @param columnFormats
	 *            column formats
	 */
	public PropertiesCSVWriter(OutputStream out, int firstIndex, Format... columnFormats) {
		this.out = out;
		this.properties = new Properties();
		this.nextIndex = firstIndex;
		this.columnFormats = columnFormats;
	}

	@Override
	public void flush() throws IOException {
	}

	@Override
	public void close() throws IOException {
		if (out != null) {
			properties.store(out, "Written by PropertiesCSVWriter");
			out.close();
		}
	}

	@Override
	public void write(String... record) throws IOException {
		if (record.length > columnFormats.length)
			throw new IOException("Too many columns");
		for (int i = 0; i < record.length; i++) {
			properties.setProperty(columnFormats[i].format(new Object[] { nextIndex }), record[i]);
		}
		nextIndex++;
	}
}
