/*
 * Copyright (c) 2010 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util.csv;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * Class to read data from an XML file as if it was a CSV (Comma Separated
 * Value) file.
 */
public class XMLCSVReader extends AbstractCSVReader {

	private Node nextRow;
	private final String rowTag;
	private final String[] cellTags;

	/**
	 * Create a new {@link XMLCSVReader} that reads from the given DOM document.
	 * All child elements of the document element are considered rows, and all
	 * their child elements are considered cells.
	 * 
	 * @param doc
	 *            the {@link Document}
	 */
	public XMLCSVReader(Document doc) {
		this(doc, "*", "*");
	}

	/**
	 * Create a new {@link XMLCSVReader} that reads from the given DOM document.
	 * All child elements of the document element that have the row tag as name
	 * (or all of them if the row tag is <code>*</code>) are considered rows,
	 * and all their child elements that have one of the cell tags as name are
	 * considered cells. In case there is more than one cell tag, the cell tag
	 * in the list of cell tags is used to determine the position in the result
	 * row. If a cell tag matches more than once in a row, all the fields before
	 * the cell tag are added to the result row, and a complete set of empty
	 * fields is used to parse the rest of the row. Cell tags that are not
	 * matched result in empty strings, unless they would be at the end of the
	 * row in the last run.
	 * 
	 * @param doc
	 *            the {@link Document}
	 * @param rowTag
	 *            the name of row elements
	 * @param cellTags
	 *            the names of cell elements
	 * 
	 */
	public XMLCSVReader(Document doc, String rowTag, String... cellTags) {
		this.rowTag = rowTag;
		this.cellTags = cellTags;
		this.nextRow = doc.getDocumentElement().getFirstChild();
	}

	/**
	 * Close this reader. This is a no-op.
	 */
	@Override
	public void close() throws IOException {
	}

	/**
	 * Read a row from this XML file.
	 */
	@Override
	public String[] read() throws IOException {
		return read(null);
	}

	/**
	 * Read a row from this XML file. If the row has a <tt>@file</tt> attribute
	 * and the fileIndex argument is non-null, the value of that file attribute
	 * is stored into the first element of that array.
	 * 
	 * @param fileIndex
	 *            Array to store the file index to, if needed
	 */
	public String[] read(int[] fileIndex) {
		while (nextRow != null && !nameMatches(nextRow, rowTag))
			nextRow = nextRow.getNextSibling();
		if (nextRow == null)
			return null;
		Element row = (Element) nextRow;
		nextRow = row.getNextSibling();
		if (row.getAttribute("file").length() > 0 && fileIndex != null)
			fileIndex[0] = Integer.parseInt(row.getAttribute("file"));
		List<String> elems = new ArrayList<String>();
		String[] restRecord = new String[cellTags.length];
		for (int i = 0; i < row.getChildNodes().getLength(); i++) {
			Node n = row.getChildNodes().item(i);
			for (int j = 0; j < cellTags.length; j++) {
				if (nameMatches(n, cellTags[j])) {
					if (restRecord[j] != null) {
						for (int k = 0; k < restRecord.length; k++) {
							elems.add(restRecord[k] == null ? "" : restRecord[k]);
							restRecord[k] = null;
						}
					}
					restRecord[j] = n.getTextContent();
					break;
				}
			}
		}
		int lastNonNull = -1;
		for (int i = 0; i < restRecord.length; i++) {
			if (restRecord[i] != null)
				lastNonNull = i;
		}
		for (int i = 0; i <= lastNonNull; i++) {
			elems.add(restRecord[i] == null ? "" : restRecord[i]);
		}
		return elems.toArray(new String[elems.size()]);
	}

	private boolean nameMatches(Node node, String name) {
		return node instanceof Element && name.equals("*") || node.getNodeName().equals(name);
	}
}
