/*
 * Copyright (c) 2010 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util.csv;

import java.io.IOException;
import java.io.OutputStream;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * Class to create an XML file as if it was a CSV (Comma Separated Value) file.
 */
public class XMLCSVWriter extends AbstractCSVWriter {

	private final OutputStream out;
	private final Node root;
	private final String rowTag;
	private final String[] cellTags;

	/**
	 * Create a new {@link XMLCSVWriter} that writes to the given DOM document.
	 * Rows will become child elements given root element that have the row tag
	 * as name, and cells will become children of the row, with the cell tags
	 * (assigned round robin) as name.
	 * 
	 * @param root
	 *            the root node
	 * @param rowTag
	 *            the name of row elements
	 * @param cellTags
	 *            the names of cell elements
	 * 
	 */
	public XMLCSVWriter(Node root, String rowTag, String... cellTags) {
		this.out = null;
		this.root = root;
		this.rowTag = rowTag;
		this.cellTags = cellTags;
	}

	/**
	 * Create a new {@link XMLCSVWriter} that writes to the given DOM document.
	 * Rows will become child elements of the document element that have the row
	 * tag as name, and cells will become children of the row, with the cell
	 * tags (assigned round robin) as name.
	 * 
	 * @param out
	 *            Output stream to write to
	 * @param rootTag
	 *            the name of the root element
	 * @param rowTag
	 *            the name of row elements
	 * @param cellTags
	 *            the names of cell elements
	 */
	public XMLCSVWriter(OutputStream out, String rootTag, String rowTag, String... cellTags) throws ParserConfigurationException {
		this.out = out;
		Document doc = DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();
		doc.appendChild(doc.createElement(rootTag));
		this.rowTag = rowTag;
		this.cellTags = cellTags;
		this.root = doc.getDocumentElement();
	}

	@Override
	public void flush() throws IOException {
	}

	@Override
	public void close() throws IOException {
		if (out != null) {
			try {
				TransformerFactory.newInstance().newTransformer().transform(new DOMSource(root.getOwnerDocument()), new StreamResult(out));
				out.close();
			} catch (Exception ex) {
				throw new IOException("Unable to save XML file", ex);
			}
		}
	}

	@Override
	public void write(String... record) throws IOException {
		Element row = root.getOwnerDocument().createElement(rowTag);
		for (int i = 0; i < record.length; i++) {
			Element cell = root.getOwnerDocument().createElement(cellTags[i % cellTags.length]);
			cell.appendChild(root.getOwnerDocument().createTextNode(record[i]));
			row.appendChild(cell);
		}
		root.appendChild(row);
	}
}
