/*
 * Copyright (c) 2010 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util.csv;

import java.io.IOException;

import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Class to read data from an XML file as if it was a CSV (Comma Separated
 * Value) file, using {@link XPathExpression}s for the columns.
 */
public class XPathCSVReader extends AbstractCSVReader {

	private final NodeList rows;
	private final XPathExpression[] columnExpressions;
	private int nextIndex = 0;

	/**
	 * Create a new {@link XPathCSVReader} that reads from the given DOM
	 * document. The rowExpression is expected to return a {@link NodeList} of
	 * rows, each columnExpression is evaluated on each of the rows to find the
	 * column values.
	 */
	public XPathCSVReader(Document doc, String rowExpression, String... columnExpressions) throws XPathExpressionException {
		this(doc, buildExpressions(rowExpression)[0], buildExpressions(columnExpressions));
	}

	/**
	 * Create a new {@link XPathCSVReader} that reads from the given DOM
	 * document. The rowExpression is expected to return a {@link NodeList} of
	 * rows, each columnExpression is evaluated on each of the rows to find the
	 * column values.
	 */
	public XPathCSVReader(Document doc, XPathExpression rowExpression, XPathExpression... columnExpressions) throws XPathExpressionException {
		this.columnExpressions = columnExpressions;
		rows = (NodeList) rowExpression.evaluate(doc, XPathConstants.NODESET);
	}

	private static final XPath xpath = XPathFactory.newInstance().newXPath();

	private static XPathExpression[] buildExpressions(String... expressions) throws XPathExpressionException {
		XPathExpression[] result = new XPathExpression[expressions.length];
		for (int i = 0; i < result.length; i++) {
			result[i] = xpath.compile(expressions[i]);
		}
		return result;
	}

	/**
	 * Close this reader. This is a no-op.
	 */
	@Override
	public void close() throws IOException {
	}

	/**
	 * Read a row from this XML file.
	 */
	@Override
	public String[] read() throws IOException {
		if (nextIndex == rows.getLength())
			return null;
		Node row = rows.item(nextIndex);
		nextIndex++;
		String[] record = new String[columnExpressions.length];
		for (int i = 0; i < record.length; i++) {
			try {
				record[i] = columnExpressions[i].evaluate(row);
			} catch (XPathExpressionException ex) {
				throw new IOException("Error while parsing " + columnExpressions[i], ex);
			}
		}
		return record;
	}
}
