/*
 * Copyright (c) 2008 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util.csv.fieldreader;

import java.util.HashMap;
import java.util.Map;

import net.sf.jmatchparser.util.csv.AbstractCSVReader;
import net.sf.jmatchparser.util.csv.fieldreader.FieldSource.EmptyAction;

/**
 * A {@link FieldReader} that uses an {@link Enum} that implements
 * {@link FieldEnum} as its field type.
 */
public class EnumFieldReader<K extends FieldEnum<V>, V> extends FieldReader<K, V> {

	private final K[] enumValues;

	/**
	 * Create a new {@link EnumFieldReader}
	 * 
	 * @param enumValues
	 *            Values of the enum to be used, usually obtained from
	 *            <code>Enum.values()</code>
	 * @param reader
	 *            CSV reader to be used
	 */
	public EnumFieldReader(K[] enumValues, AbstractCSVReader reader) {
		super(buildDefinitions(enumValues), reader);
		this.enumValues = enumValues;
	}

	private static <K extends FieldEnum<V>, V> Map<K, FieldDefinition<? extends V>> buildDefinitions(K[] enumValues) {
		Map<K, FieldDefinition<? extends V>> result = new HashMap<K, FieldDefinition<? extends V>>();
		for (K enumValue : enumValues) {
			result.put(enumValue, enumValue.getDefinition());
		}
		return result;
	}

	/**
	 * Compute the field sources from the index of the enum value inside the
	 * declaration.
	 * 
	 * To be used when header parsing is disabled.
	 * 
	 * @param emptyAction
	 *            {@link EmptyAction} to use
	 */
	public void computeFieldSourcesFromIndex(EmptyAction emptyAction) {
		computeFieldSourcesFromIndex(emptyAction, enumValues);
	}
}
