/*
 * Copyright (c) 2008 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util.csv.fieldreader;

import java.text.ParseException;
import java.util.Map;
import java.util.regex.Pattern;

import net.sf.jmatchparser.util.PatternCache;
import net.sf.jmatchparser.util.csv.fieldreader.FieldReaderException.FieldReaderExceptionInfo;

/**
 * A definition of a field of the {@link FieldReader}, as seen from the
 * perspective of the "consumer" that uses the parsed data.
 * 
 * @param <T>
 *            Type of the parsed object
 */
public class FieldDefinition<T> {

	private final Pattern mask;
	private final int maxLength;
	private final ParseFormat<T> format;
	private final MandatoryState mandatory;

	/**
	 * Create a new field definition.
	 * 
	 * @param mask
	 *            Regular expression the values have to match before being
	 *            formatted.
	 * @param maxLength
	 *            Maximum length of the values before being formatted
	 * @param mandatory
	 *            mandatory state of this field (see {@link MandatoryState})
	 * @param format
	 *            Format to be applied to the value
	 */
	public FieldDefinition(Pattern mask, int maxLength, MandatoryState mandatory, ParseFormat<T> format) {
		this.mask = mask;
		this.maxLength = maxLength;
		this.mandatory = mandatory;
		this.format = format;
	}

	/**
	 * Create a new field definition.
	 * 
	 * @param mask
	 *            Regular expression the values have to match before being
	 *            formatted.
	 * @param maxLength
	 *            Maximum length of the values before being formatted
	 * @param mandatory
	 *            mandatory state of this field (see {@link MandatoryState})
	 * @param format
	 *            Format to be applied to the value
	 */
	public FieldDefinition(String mask, int maxLength, MandatoryState mandatory, ParseFormat<T> format) {
		this(PatternCache.compile(mask), maxLength, mandatory, format);
	}

	MandatoryState getMandatory() {
		return mandatory;
	}

	T parse(String value, String fieldName, Map<String, String> rawRecord) throws FieldReaderException {
		if (maxLength < value.length()) {
			throw new FieldReaderException(FieldReaderErrorCodes.VALUE_TOO_LONG, true, new FieldReaderExceptionInfo<String, String>(fieldName, value, rawRecord));
		}
		if (!mask.matcher(value).matches()) {
			throw new FieldReaderException(FieldReaderErrorCodes.INVALID_FORMAT, true, new FieldReaderExceptionInfo<String, String>(fieldName, value, rawRecord));
		}
		try {
			return format.parse(value);
		} catch (ParseException ex) {
			throw new FieldReaderException(FieldReaderErrorCodes.INVALID_FORMAT, true, new FieldReaderExceptionInfo<String, String>(fieldName, value, rawRecord), ex);
		}
	}

	/**
	 * Mandatory state of a field.
	 */
	public enum MandatoryState {

		/**
		 * This field is optional. It can be filled or not.
		 */
		OPTIONAL,

		/**
		 * This field is mandatory. It has to be filled for all records.
		 */
		MANDATORY,

		/**
		 * This field is only mandatory if a new record should be created, not
		 * if a record should be updated.
		 */
		MANDATORY_IF_NEW
	}
}
