/*
 * Copyright (c) 2008 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util.csv.fieldreader;

import java.io.IOException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import net.sf.jmatchparser.util.csv.AbstractCSVReader;
import net.sf.jmatchparser.util.csv.fieldreader.FieldDefinition.MandatoryState;
import net.sf.jmatchparser.util.csv.fieldreader.FieldReaderException.FieldReaderExceptionInfo;
import net.sf.jmatchparser.util.csv.fieldreader.FieldSource.EmptyAction;

/**
 * A parser that can be used to parse different CSV files into the same set of
 * records. It is useful for building parsers that need a fixed range of
 * information, but the files that contain them can have a multitude of
 * different formats; often the parser that consumes the records of this field
 * reader (who supplies {@link FieldDefinition}s) is in a different library or
 * component than the creator of this field reader (who supplies
 * {@link FieldSource}s).
 * 
 * <p>
 * Each record is formatted by this field reader into a {@link Map} that maps
 * fields (usually either {@link String}s or an {@link Enum} that implements
 * {@link FieldEnum}, see {@link EnumFieldReader}) to values (sometimes strings,
 * but often other objects as well).
 * 
 * <p>
 * By convention, keys that are mapped to <code>null</code> are expected to be
 * cleared by the parser (when updating anything), and records that are not
 * present are expected not to be touched. Therefore, parsers that use this
 * reader should use the {@link Map#containsKey(Object)} method to determine if
 * a key is present, instead of checking the value of {@link Map#get(Object)}
 * for <code>null</code>. Alternatively, the {@link #setEmptyFieldValue(Object)}
 * method can be used to set the "empty" (clear) value to something else.
 * 
 * @param <K>
 *            The type of the fields
 * @param <V>
 *            The type of the parsed values
 */
public class FieldReader<K, V> {

	private final Map<K, FieldDefinition<? extends V>> fieldDefinitions;
	private final AbstractCSVReader reader;
	private Map<K, FieldSource> fieldSources = null;
	private V emptyFieldValue = null;
	private String[] header = null;

	/**
	 * Create a new field reader.
	 * 
	 * @param fieldDefinitions
	 *            Field definitions to use
	 * @param reader
	 *            Reader to read the records from
	 */
	public FieldReader(Map<K, FieldDefinition<? extends V>> fieldDefinitions, AbstractCSVReader reader) {
		this.fieldDefinitions = fieldDefinitions;
		this.reader = reader;
	}

	/**
	 * Set the {@link FieldSource}s to be used for this field reader.
	 */
	public void setFieldSources(Map<K, FieldSource> fieldSources) {
		if (this.fieldSources != null)
			throw new IllegalStateException("Field sources have already been assigned");
		this.fieldSources = fieldSources;
		for (K key : fieldSources.keySet()) {
			if (!fieldDefinitions.containsKey(key)) {
				throw new IllegalStateException("Field source for field without definition: " + key);
			}
		}
	}

	/**
	 * Compute the field sources automatically from the index of the key inside
	 * the given array.
	 * 
	 * To be used when header parsing is disabled.
	 * 
	 * @param emptyAction
	 *            {@link EmptyAction} to use
	 * @param fields
	 *            Fields in the desired order
	 */
	public void computeFieldSourcesFromIndex(EmptyAction emptyAction, K[] fields) {
		Map<K, FieldSource> fieldSources = new HashMap<K, FieldSource>();
		for (int i = 0; i < fields.length; i++) {
			fieldSources.put(fields[i], FieldSource.withChecks(FieldSource.fromColumn("" + (i + 1)), emptyAction, ".*"));
		}
		setFieldSources(fieldSources);
	}

	/**
	 * Compute the field sources automatically from the names (see
	 * {@link #toString()}) of all fields
	 * 
	 * To be used when header parsing is enabled.
	 * 
	 * @param emptyAction
	 *            {@link EmptyAction} to use
	 */
	public void computeFieldSourcesFromName(EmptyAction emptyAction) {
		computeFieldSourcesFromName(emptyAction, fieldDefinitions.keySet());
	}

	/**
	 * Compute the field sources automatically from the names (see
	 * {@link #toString()}) of the given fields
	 * 
	 * To be used when header parsing is enabled.
	 * 
	 * @param emptyAction
	 *            {@link EmptyAction} to use
	 * @param fields
	 *            Fields to use
	 */
	public void computeFieldSourcesFromName(EmptyAction emptyAction, Collection<K> fields) {
		Map<K, FieldSource> fieldSources = new HashMap<K, FieldSource>();
		for (K key : fields) {
			fieldSources.put(key, FieldSource.withChecks(FieldSource.fromColumn(key.toString()), emptyAction, ".*"));
		}
		setFieldSources(fieldSources);
	}

	/**
	 * Read a header line from the CSV reader, and optionally use it for column
	 * assignment.
	 * 
	 * @param ignore
	 *            Whether to ignore the read header (not use it for column
	 *            assignment)
	 */
	public void readHeader(boolean ignore) throws IOException {
		String[] header = reader.read();
		if (!ignore)
			this.header = header;
	}

	/**
	 * Set the value that should be used for records that should be updated, but
	 * to an empty value (by default <code>null</code>).
	 * 
	 * @param emptyFieldValue
	 *            new value for empty fields
	 */
	public void setEmptyFieldValue(V emptyFieldValue) {
		this.emptyFieldValue = emptyFieldValue;
	}

	/**
	 * Read a record from this field reader.
	 */
	public Map<K, V> read() throws IOException, FieldReaderException {
		if (fieldSources == null)
			throw new IllegalStateException("Field sources have not been assigned");
		String[] record = reader.read();

		if (record == null)
			return null;

		Map<String, String> rawRecord = new HashMap<String, String>();
		for (int i = 0; i < record.length; i++) {
			if (record[i].length() == 0)
				continue;
			String key = (header != null && i < header.length) ? header[i] : "" + (i + 1);
			rawRecord.put(key, record[i]);
		}
		Map<K, V> result = new HashMap<K, V>();
		for (Map.Entry<K, FieldSource> sourceEntry : fieldSources.entrySet()) {
			K key = sourceEntry.getKey();
			FieldSource source = sourceEntry.getValue();
			FieldDefinition<? extends V> definition = fieldDefinitions.get(key);
			String stringValue = source.getValue(rawRecord);
			if (stringValue == null)
				continue;
			if (stringValue.length() == 0) {
				result.put(key, emptyFieldValue);
			} else {
				result.put(key, definition.parse(stringValue, source.getField(), rawRecord));
			}
		}
		for (Map.Entry<K, FieldDefinition<? extends V>> definitionEntry : fieldDefinitions.entrySet()) {
			K key = definitionEntry.getKey();
			FieldDefinition<? extends V> definition = definitionEntry.getValue();
			if (!result.containsKey(key) && definition.getMandatory() == MandatoryState.MANDATORY) {
				throw new FieldReaderException(FieldReaderErrorCodes.MISSING, false, new FieldReaderExceptionInfo<K, V>(key, null, result));
			}
		}
		return result;
	}

	/**
	 * Verify and postprocess a new record (i. e. one that does not already
	 * exist in the underlying database and has to be created instead of
	 * updated.
	 * 
	 * Note that the current implementation does not perform any postprocessing
	 * and returns the record as is.
	 * 
	 * @param record
	 *            Record that has been read by this field reader before
	 * @return postprocessed record
	 * @throws FieldReaderException
	 *             if mandatory fields are missing
	 */
	public Map<K, V> postprocessNewRecord(Map<K, V> record) throws FieldReaderException {
		for (Map.Entry<K, FieldSource> sourceEntry : fieldSources.entrySet()) {
			K key = sourceEntry.getKey();
			FieldSource source = sourceEntry.getValue();
			if (record.containsKey(key))
				continue;
			if (source.isMandatoryIfNew()) {
				throw new FieldReaderException(FieldReaderErrorCodes.MISSING_MANDATORYIFNEW, false, new FieldReaderExceptionInfo<K, V>(key, null, record));
			}
		}
		for (Map.Entry<K, FieldDefinition<? extends V>> definitionEntry : fieldDefinitions.entrySet()) {
			K key = definitionEntry.getKey();
			FieldDefinition<? extends V> definition = definitionEntry.getValue();
			if (record.containsKey(key))
				continue;
			if (definition.getMandatory() == MandatoryState.MANDATORY_IF_NEW) {
				throw new FieldReaderException(FieldReaderErrorCodes.MISSING_MANDATORYIFNEW, false, new FieldReaderExceptionInfo<K, V>(key, null, record));
			}
		}
		return record;
	}

	/**
	 * Close this field reader.
	 */
	public void close() throws IOException {
		reader.close();
	}
}
