/*
 * Copyright (c) 2008 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util.csv.fieldreader;

import java.util.Map;

/**
 * Thrown when a field has a value that cannot be parsed.
 */
public class FieldReaderException extends Exception {

	private final FieldReaderErrorCode errorCode;
	private final FieldReaderExceptionInfo<?, ?> exceptionInfo;
	private final boolean rawField;

	/**
	 * Create a new field reader exception.
	 * 
	 * @param errorCode
	 *            Error code
	 * @param rawField
	 *            Whether the field is still raw (like in the CSV layout) or
	 *            already formatted (in the layout the {@link FieldReader}
	 *            creates)
	 * @param exceptionInfo
	 *            Information about the field and its value
	 */
	public FieldReaderException(FieldReaderErrorCode errorCode, boolean rawField, FieldReaderExceptionInfo<?, ?> exceptionInfo) {
		this(errorCode, rawField, exceptionInfo, null);
	}

	/**
	 * Create a new field reader exception.
	 * 
	 * @param errorCode
	 *            Error code
	 * @param rawField
	 *            Whether the field is still raw (like in the CSV layout) or
	 *            already formatted (in the layout the {@link FieldReader}
	 *            creates)
	 * @param exceptionInfo
	 *            Information about the field and its value
	 */
	public FieldReaderException(FieldReaderErrorCode errorCode, boolean rawField, FieldReaderExceptionInfo<?, ?> exceptionInfo, Throwable cause) {
		super((rawField ? "Raw " : "") + "Field " + exceptionInfo.getField() + ", Value: " + exceptionInfo.getValue() + ", Error: "
				+ errorCode.getErrorCode() + " (" + errorCode.getErrorDescription() + "), Record: "
				+ exceptionInfo.getOriginalRecord().toString(), cause);
		this.errorCode = errorCode;
		this.rawField = rawField;
		this.exceptionInfo = exceptionInfo;
	}

	/**
	 * Return the error code of this exception.
	 */
	public FieldReaderErrorCode getErrorCode() {
		return errorCode;
	}

	/**
	 * Return whether this exception occurred when the fields still were "raw"
	 * (i. e. not formatted).
	 */
	public boolean isRawField() {
		return rawField;
	}

	/**
	 * Return the exception info of this exception.
	 */
	public FieldReaderExceptionInfo<?, ?> getExceptionInfo() {
		return exceptionInfo;
	}

	/**
	 * Additional information about a field reader exception. Placed into a
	 * nested class since {@link Exception}s may not have generic arguments.
	 * 
	 * @param <K>
	 *            Type of the fields of the field reader
	 * @param <V>
	 *            Type of the values of the field reader
	 */
	public static class FieldReaderExceptionInfo<K, V> {
		private final K field;
		private final V value;
		private final Map<K, V> originalRecord;

		/**
		 * Create a new FieldReaderExceptionInfo instance.
		 * 
		 * @param field
		 *            Field that caused the exception
		 * @param value
		 *            Value of that field
		 * @param originalRecord
		 *            All fields as available at that state of parsing
		 */
		public FieldReaderExceptionInfo(K field, V value, Map<K, V> originalRecord) {
			this.field = field;
			this.value = value;
			this.originalRecord = originalRecord;
		}

		/**
		 * Return the field.
		 */
		public K getField() {
			return field;
		}

		/**
		 * Return the value.
		 */
		public V getValue() {
			return value;
		}

		/**
		 * Return all the other fields available.
		 */
		public Map<K, V> getOriginalRecord() {
			return originalRecord;
		}
	}
}
