/*
 * Copyright (c) 2008 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util.csv.fieldreader;

import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import net.sf.jmatchparser.util.PatternCache;
import net.sf.jmatchparser.util.csv.fieldreader.FieldReaderException.FieldReaderExceptionInfo;

/**
 * A description of the "source" from where data should be filled into the
 * field, as seen from the perspective of the "provider" that provides the CSV
 * files.
 */
public abstract class FieldSource {

	/**
	 * Return the value of this field source.
	 * 
	 * @param columns
	 *            Hashtable of fields read from the CSV file
	 * @return Value, or <code>null</code> if the field should be treated as
	 *         absent
	 */
	public abstract String getValue(Map<String, String> columns) throws FieldReaderException;

	/**
	 * Return an artificial name for this field, used for debugging and in error
	 * messages.
	 */
	public abstract String getField();

	/**
	 * Return whether this field has to be filled for new records
	 */
	public boolean isMandatoryIfNew() {
		return false;
	}

	/**
	 * Create a new field source that always returns the same constant value.
	 * 
	 * @param constant
	 *            The constant value
	 */
	public static FieldSource fromConstant(final String constant) {
		return new FieldSource() {
			@Override
			public String getValue(Map<String, String> columns) {
				return constant;
			}

			@Override
			public String getField() {
				return constant;
			}
		};
	}

	/**
	 * Create a new field source that returns a value from a CSV column.
	 * 
	 * @param columnName
	 *            Name or index of the column
	 */
	public static FieldSource fromColumn(final String columnName) {
		return new FieldSource() {
			@Override
			public String getValue(Map<String, String> columns) {
				String result = columns.get(columnName);
				return result == null ? "" : result;
			}

			@Override
			public String getField() {
				return "{" + columnName + "}";
			}
		};
	}

	/**
	 * Create a new field source that returns a value concatenated from several
	 * other {@link FieldSource}s.
	 * 
	 * @param sources
	 *            Other field sources to be concatenated
	 */
	public static FieldSource concat(final FieldSource... sources) {
		return new FieldSource() {
			@Override
			public String getValue(Map<String, String> columns) throws FieldReaderException {
				StringBuilder sb = new StringBuilder();
				for (FieldSource source : sources) {
					sb.append(source.getValue(columns));
				}
				return sb.toString();
			}

			@Override
			public String getField() {
				StringBuilder sb = new StringBuilder();
				for (FieldSource source : sources) {
					sb.append(source.getField());
				}
				return sb.toString();
			}
		};
	}

	/**
	 * Create a field source that performs a series of regex replacements on
	 * another field source.
	 * 
	 * @param source
	 *            Field source
	 * @param mappings
	 *            A list delimited by <tt>;;</tt> of regex replacement pairs,
	 *            where <tt>::</tt> separates regex and replacement, or
	 *            <code>null</code> for no mappings
	 */
	public static FieldSource withMappings(FieldSource source, String mappings) {
		if (mappings == null)
			return source;
		final String[] mappingArray = mappings.split(";;");
		for (int i = mappingArray.length - 1; i >= 0; i--) {
			String[] parts = mappingArray[i].split("::", 2);
			if (parts.length == 2) {
				source = withMapping(source, parts[0], parts[1]);
			}
		}
		return source;
	}

	/**
	 * Create a field source that performs a regex replacement on another field
	 * source.
	 * 
	 * @param source
	 *            Field source
	 * @param regex
	 *            regular expression
	 * @param replacement
	 *            replacement string
	 */
	public static FieldSource withMapping(FieldSource source, String regex, String replacement) {
		return withMapping(source, PatternCache.compile(regex), replacement);
	}

	/**
	 * Create a field source that performs a regex replacement on another field
	 * source.
	 * 
	 * @param source
	 *            Field source
	 * @param regex
	 *            regular expression
	 * @param replacement
	 *            replacement string
	 */
	public static FieldSource withMapping(FieldSource source, Pattern regex, String replacement) {
		return withFormat(source, ParseFormat.fromReplacement(regex, replacement, ParseFormat.IDENTITY));
	}

	/**
	 * Create a field source that applies a {@link ParseFormat} on another field
	 * source.
	 * 
	 * @param source
	 *            Field source
	 * @param format
	 *            {@link ParseFormat}
	 */
	public static FieldSource withFormat(final FieldSource source, final ParseFormat<String> format) {
		return new FieldSource() {

			@Override
			public String getValue(Map<String, String> columns) throws FieldReaderException {
				String value = source.getValue(columns);
				try {
					return format.parse(value);
				} catch (ParseException ex) {
					throw new FieldReaderException(FieldReaderErrorCodes.INVALID_FORMAT, true, new FieldReaderExceptionInfo<String, String>(source.getField(), value, columns), ex);
				}
			}

			@Override
			public String getField() {
				return source.getField();
			}
		};
	}

	/**
	 * Create a field source that performs checks on the output of another field
	 * source and returns the result verbatim. This field source is usually the
	 * outermost field source used in a stack of other field sources.
	 * 
	 * @param source
	 *            field source
	 * @param emptyAction
	 *            Action to perform when the field source returns an empty
	 *            string
	 * @param mask
	 *            regex pattern the field source result has to match, or
	 *            <code>null</code>
	 */
	public static FieldSource withChecks(FieldSource source, EmptyAction emptyAction, String mask) {
		return withChecks(source, emptyAction, mask == null ? null : PatternCache.compile(mask));
	}

	/**
	 * Create a field source that performs checks on the output of another field
	 * source and returns the result verbatim. This field source is usually the
	 * outermost field source used in a stack of other field sources.
	 * 
	 * @param source
	 *            field source
	 * @param emptyAction
	 *            Action to perform when the field source returns an empty
	 *            string
	 * @param mask
	 *            regex pattern the field source result has to match, or
	 *            <code>null</code>
	 */
	public static FieldSource withChecks(final FieldSource source, final EmptyAction emptyAction, final Pattern mask) {
		return new FieldSource() {

			@Override
			public String getValue(Map<String, String> columns) throws FieldReaderException {
				String value = source.getValue(columns);

				if (value.length() == 0) {
					if (emptyAction == EmptyAction.COMPLAIN) {
						throw new FieldReaderException(FieldReaderErrorCodes.MISSING, true, new FieldReaderExceptionInfo<String, String>(source.getField(), "", columns));
					} else if (emptyAction != EmptyAction.CLEAR) {
						value = null;
					}
				}
				if (value != null && mask != null
						&& !mask.matcher(value).matches()) {
					throw new FieldReaderException(FieldReaderErrorCodes.INVALID_FORMAT, true, new FieldReaderExceptionInfo<String, String>(source.getField(), value, columns));
				}

				return value;
			}

			@Override
			public boolean isMandatoryIfNew() {
				return emptyAction == EmptyAction.COMPLAIN_IF_NEW;
			}

			@Override
			public String getField() {
				return source.getField();
			}
		};
	}

	/**
	 * Parse a field source from several parameters.
	 * 
	 * @param columns
	 *            Columns spec, either one or multiple column names separated by
	 *            one of <tt>|,;: </tt> (the separator will be added to the
	 *            result), or a literal string where column names are included
	 *            in curly braces
	 * @param emptyAction
	 *            Action to perform when the field source returns an empty
	 *            string
	 * @param mask
	 *            regex pattern the field source result has to match
	 * @param mappings
	 *            A list delimited by <tt>;;</tt> of regex replacement pairs,
	 *            where <tt>::</tt> separates regex and replacement.
	 * 
	 * @see #withChecks(FieldSource, EmptyAction, Pattern)
	 * @see #withMappings(FieldSource, String)
	 */
	public static FieldSource parse(String columns, EmptyAction emptyAction, String mask, String mappings) {
		FieldSource[] sources;
		if (columns.indexOf('{') == -1)
			sources = parseLegacyColumns(columns);
		else
			sources = parseColumns(columns);
		FieldSource result = concat(sources);
		result = withMappings(result, mappings);
		return withChecks(result, emptyAction, mask);
	}

	private static FieldSource[] parseColumns(String columns) {
		Matcher m = PatternCache.compile("\\{(.*?)\\}").matcher(columns);
		List<FieldSource> result = new ArrayList<FieldSource>();
		StringBuffer constantBuffer = new StringBuffer();
		while (m.find()) {
			m.appendReplacement(constantBuffer, "");
			if (constantBuffer.length() > 0) {
				result.add(fromConstant(constantBuffer.toString()));
				constantBuffer.setLength(0);
			}
			// workaround to be able to use fully constant strings by adding {}
			// somewhere in them
			if (m.group(1).length() > 0) {
				result.add(fromColumn(m.group(1)));
			}
		}
		m.appendTail(constantBuffer);
		if (constantBuffer.length() > 0) {
			result.add(fromConstant(constantBuffer.toString()));
			constantBuffer.setLength(0);
		}
		return result.toArray(new FieldSource[result.size()]);
	}

	private static FieldSource[] parseLegacyColumns(String columns) {
		final StringTokenizer st = new StringTokenizer(columns, "|,;: ", true);
		final FieldSource[] result = new FieldSource[st.countTokens()];
		for (int i = 0; i < result.length; i++) {
			String token = st.nextToken();
			if (token.length() == 1 && "|,;: ".indexOf(token) != -1) {
				result[i] = fromConstant(token);
			} else {
				result[i] = fromColumn(token);
			}
		}
		return result;
	}

	/**
	 * How to treat a field source that returns an empty string.
	 */
	public static enum EmptyAction {

		/**
		 * Ignore the field as if it was not in the field source list, to not
		 * touch the underlying field when updating a record.
		 */
		IGNORE,

		/**
		 * Throw an exception.
		 */
		COMPLAIN,

		/**
		 * Return an empty value, to clear the underlying field when updating a
		 * record.
		 */
		CLEAR,

		/**
		 * Throw an exception if a new record should be created, ignore it
		 * otherwise.
		 * 
		 * @see #COMPLAIN
		 * @see #IGNORE
		 */
		COMPLAIN_IF_NEW;
	}
}
