/*
 * Copyright (c) 2008 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util.csv.fieldreader;

import java.io.IOException;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import net.sf.jmatchparser.util.csv.AbstractCSVReader;
import net.sf.jmatchparser.util.csv.fieldreader.FieldDefinition.MandatoryState;
import net.sf.jmatchparser.util.csv.fieldreader.FieldReaderException.FieldReaderExceptionInfo;
import net.sf.jmatchparser.util.csv.fieldreader.FieldSource.EmptyAction;

/**
 * Class to create field sources from another CSV file.
 */
public class FieldSourceFactory {

	/**
	 * Fields used in the CSV file for defining field sources
	 */
	public enum FieldSourceFactoryFields implements FieldEnum<Object> {

		/**
		 * Name of the field to be defined.
		 */
		FIELD(new SimpleFieldDefinition(Integer.MAX_VALUE, MandatoryState.MANDATORY)),

		/**
		 * Columns to be used, in the same format as used by
		 * {@link FieldSource#parse(String, EmptyAction, String, String)}.
		 */
		COLUMNS(new SimpleFieldDefinition(Integer.MAX_VALUE, MandatoryState.MANDATORY)),

		/**
		 * {@link EmptyAction} to be used.
		 */
		EMPTYACTION(new FieldDefinition<EmptyAction>("[A-Z_]+", 15, MandatoryState.MANDATORY, ParseFormat.forEnum(EmptyAction.class))),

		/**
		 * Regex mask.
		 */
		MASK(new SimpleFieldDefinition()),

		/**
		 * Regex mapping, in the same format as used by
		 * {@link FieldSource#withMappings(FieldSource, String)}.
		 */
		MAPPINGS(new SimpleFieldDefinition());

		private final FieldDefinition<? extends Object> definition;

		private FieldSourceFactoryFields(FieldDefinition<? extends Object> definition) {
			this.definition = definition;
		}

		@Override
		public FieldDefinition<? extends Object> getDefinition() {
			return definition;
		}
	}

	private FieldReader<FieldSourceFactoryFields, Object> reader;

	/**
	 * Create a new field source factory that reads its field sources from the
	 * given CSV reader.
	 */
	public FieldSourceFactory(AbstractCSVReader reader) {
		this.reader = new EnumFieldReader<FieldSourceFactory.FieldSourceFactoryFields, Object>(FieldSourceFactoryFields.values(), reader);
		this.reader.setEmptyFieldValue("");
	}

	/**
	 * Get the {@link FieldReader} used to parse the CSV file, for customizing
	 * its field sources or other parameters.
	 */
	public FieldReader<FieldSourceFactoryFields, Object> getReader() {
		return reader;
	}

	/**
	 * Create field sources from the CSV file.
	 * 
	 * @param <K>
	 *            Type of the fields used
	 * @param possibleFields
	 *            Array of all possible fields, used to convert to string via
	 *            {@link #toString()} and compare
	 */
	public <K> Map<K, FieldSource> createFieldSources(K[] possibleFields) throws FieldReaderException, IOException {
		return createFieldSources(Arrays.asList(possibleFields));
	}

	/**
	 * Create field sources from the CSV file.
	 * 
	 * @param <K>
	 *            Type of the fields used
	 * @param possibleFields
	 *            All possible fields, used to convert to string via
	 *            {@link #toString()} and compare
	 */
	public <K> Map<K, FieldSource> createFieldSources(Collection<K> possibleFields) throws FieldReaderException, IOException {
		Map<String, K> fieldsByName = new HashMap<String, K>();
		for (K field : possibleFields) {
			fieldsByName.put(field.toString(), field);
		}
		Map<K, FieldSource> result = new HashMap<K, FieldSource>();
		Map<FieldSourceFactoryFields, Object> record;
		while ((record = reader.read()) != null) {
			String field = (String) record.get(FieldSourceFactoryFields.FIELD);
			String columns = (String) record.get(FieldSourceFactoryFields.COLUMNS);
			EmptyAction emptyAction = (EmptyAction) record.get(FieldSourceFactoryFields.EMPTYACTION);
			String mask = (String) record.get(FieldSourceFactoryFields.MASK);
			String mappings = (String) record.get(FieldSourceFactoryFields.MAPPINGS);
			if (!fieldsByName.containsKey(field)) {
				throw new FieldReaderException(FieldReaderErrorCodes.INVALID_FORMAT, false, new FieldReaderExceptionInfo<FieldSourceFactoryFields, Object>(FieldSourceFactoryFields.FIELD, field, record));
			}
			result.put(fieldsByName.get(field), FieldSource.parse(columns, emptyAction, mask, mappings));
		}
		reader.close();
		return result;
	}
}
