/*
 * Copyright (c) 2008 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util.csv.fieldreader;

import java.text.DateFormat;
import java.text.Format;
import java.text.NumberFormat;
import java.text.ParseException;
import java.text.ParsePosition;
import java.util.Date;
import java.util.regex.Pattern;

/**
 * A format for parsing a String to an object of a possibly different type.
 * 
 * @param <T>
 *            Type of the parsed object
 */
public abstract class ParseFormat<T> {

	/**
	 * Parse the given source completely. An exception is thrown if not the
	 * complete source has been parsed.
	 * 
	 * @param source
	 *            Source string
	 * @return Parsed object
	 */
	public abstract T parse(String source) throws ParseException;

	/**
	 * Parse format that returns the original string.
	 */
	public static final ParseFormat<String> IDENTITY = new ParseFormat<String>() {
		@Override
		public String parse(String source) throws ParseException {
			return source;
		}
	};

	/**
	 * Parse format that uses a {@link Format} for parsing.
	 */
	public static ParseFormat<Object> fromFormat(final Format format) {
		return new ParseFormat<Object>() {
			@Override
			public Object parse(String source) throws ParseException {
				ParsePosition pp = new ParsePosition(0);
				Object result = format.parseObject(source, pp);
				checkParsePosition(source, pp);
				return result;
			}
		};
	}

	/**
	 * Parse format that uses a {@link DateFormat} for parsing.
	 */
	public static ParseFormat<Date> fromDateFormat(final DateFormat format) {
		return new ParseFormat<Date>() {
			@Override
			public Date parse(String source) throws ParseException {
				ParsePosition pp = new ParsePosition(0);
				Date result = format.parse(source, pp);
				checkParsePosition(source, pp);
				return result;
			}
		};
	}

	/**
	 * Parse format that uses a {@link NumberFormat} for parsing.
	 */
	public static ParseFormat<Number> fromNumberFormat(final NumberFormat format) {
		return new ParseFormat<Number>() {
			@Override
			public Number parse(String source) throws ParseException {
				ParsePosition pp = new ParsePosition(0);
				Number result = format.parse(source, pp);
				checkParsePosition(source, pp);
				return result;
			}
		};
	}

	/**
	 * Parse format that performs a regex replacement before parsing with
	 * another format.
	 */
	public static <TT> ParseFormat<TT> fromReplacement(final Pattern regex, final String replacement, final ParseFormat<TT> nextFormat) {
		return new ParseFormat<TT>() {
			@Override
			public TT parse(String source) throws ParseException {
				return nextFormat.parse(regex.matcher(source).replaceAll(replacement));
			}
		};
	}

	/**
	 * Parse format that parses {@link Enum} names to their values.
	 */
	public static <TT extends Enum<TT>> ParseFormat<TT> forEnum(final Class<TT> enumType) {
		return new ParseFormat<TT>() {
			@Override
			public TT parse(String source) throws ParseException {
				try {
					return Enum.valueOf(enumType, source);
				} catch (IllegalArgumentException ex) {
					ParseException pex = new ParseException("Unable to parse enum", 0);
					pex.initCause(ex);
					throw pex;
				}
			}
		};
	}

	private static void checkParsePosition(String source, ParsePosition pp) throws ParseException {
		if (pp.getIndex() != source.length())
			throw new ParseException("Parsing stopped before end of string", pp.getErrorIndex() == -1 ? pp.getIndex() : pp.getErrorIndex());
	}
}
