/*
 * Copyright (c) 2006 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package net.sf.jmatchparser.util.split;

/**
 * A split rule that creates a new output file every given number of bytes or
 * lines.
 */
public class CountSplitRule extends SplitRule {

	private final int desiredSize;
	private int current;
	private final char charToCount;

	/**
	 * Create a new rule that splits after a given number of bytes or lines.
	 * 
	 * @param desiredSize
	 *            Number of bytes/lines
	 * @param lines
	 *            <code>true</code> for lines, <code>false</code> for bytes
	 */
	public CountSplitRule(int desiredSize, boolean lines) {
		this(desiredSize, lines ? '\n' : 0);
	}

	/**
	 * Create a new rule that splits after a given number a give character.
	 * 
	 * @param desiredSize
	 *            Number of bytes/lines
	 * @param charToCount
	 *            character that should be counted, or 0 to count all characters
	 */
	public CountSplitRule(int desiredSize, char charToCount) {
		this.desiredSize = desiredSize;
		this.charToCount = charToCount;
		current = 0;
	}

	private int count(byte[] data, int off, int len) {
		if (charToCount != 0) {
			int cnt = 0;
			for (int i = off; i < off + len; i++) {
				if (data[i] == charToCount)
					cnt++;
			}
			return cnt;
		} else {
			return len;
		}
	}

	@Override
	public int findSplitPos(byte[] data, int off, int len) {
		int remaining = desiredSize - current;
		if (len < remaining)
			return -1;
		if (charToCount == 0)
			return remaining;
		for (int i = off; i < off + len; i++) {
			if (data[i] == charToCount)
				remaining--;
			if (remaining == 0)
				return i - off + 1;
		}
		return -1;
	}

	@Override
	public void update(byte[] data, int off, int len) {
		current += count(data, off, len);
		if (current > desiredSize)
			throw new IllegalStateException("SplitOutputStream has to split this!");
		if (current == desiredSize)
			current = 0;
	}
}
