/*
 * Copyright (c) 2006 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util.split;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.OutputStream;
import java.text.Format;
import java.text.MessageFormat;

/**
 * Output stream factory that creates file output streams.
 */
public class FileOutputStreamFactory implements SplitOutputStream.StreamFactory {

	private final File directory;
	private final Format filenameFormat;
	private final int firstNumber;

	/**
	 * Create a new file output stream factory that creates files in a given
	 * directory in a given format.
	 * 
	 * @param directory
	 *            Directory to put the files into.
	 * @param filenameFormat
	 *            Format string used to format the file name, as defined by
	 *            {@link MessageFormat}
	 * @param firstNumber
	 *            Number to use for the first stream
	 */
	public FileOutputStreamFactory(File directory, String filenameFormat, int firstNumber) {
		this(directory, new MessageFormat(filenameFormat), firstNumber);
	}

	/**
	 * Create a new file output stream factory that creates files in a given
	 * directory in a given format.
	 * 
	 * @param directory
	 *            Directory to put the files into.
	 * @param filenameFormat
	 *            {@link Format} used to format the file name
	 * @param firstNumber
	 *            Number to use for the first stream
	 */
	public FileOutputStreamFactory(File directory, Format filenameFormat, int firstNumber) {
		this.directory = directory;
		this.filenameFormat = filenameFormat;
		this.firstNumber = firstNumber;
	}

	@Override
	public OutputStream createOutputStream(int index) throws FileNotFoundException {
		String filename = filenameFormat.format(new Object[] { new Integer(index + firstNumber) });
		return new FileOutputStream(new File(directory, filename));
	}
}
