/*
 * Copyright (c) 2006 - 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util.split;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;

/**
 * Output stream that can split the output into multiple streams/files, and may
 * optionally add a header and/or footer to each stream. A {@link SplitRule} is
 * used to determine the split positions.
 */
public class SplitOutputStream extends OutputStream {

	private OutputStream out = null;
	private OutMode outMode = OutMode.START;
	private final StreamFactory factory;
	private final SplitRule rule;
	private byte[] header = null, footer = null;
	private int index = 0;

	/**
	 * Create a new split output stream.
	 * 
	 * @param factory
	 *            Factory used to create streams
	 * @param rule
	 *            Split rule to determine when to split the stream
	 */
	public SplitOutputStream(StreamFactory factory, SplitRule rule) {
		this.factory = factory;
		this.rule = rule;
	}

	@Override
	public void write(byte[] b, int off, int len) throws IOException {
		if (shouldCount()) {
			int splitPos = rule.findSplitPos(b, off, len);
			if (splitPos == -1) {
				out.write(b, off, len);
				rule.update(b, off, len);
			} else if (splitPos == len) {
				out.write(b, off, len);
				rule.update(b, off, len);
				next();
			} else if (len == 1) {
				throw new IOException("Cannot split one byte");
			} else {
				// recursive, divide&conquer
				write(b, off, splitPos);
				write(b, off + splitPos, len - splitPos);
			}
		} else {
			out.write(b, off, len);
		}
	}

	@Override
	public void write(int b) throws IOException {
		if (shouldCount()) {
			int splitPos = rule.findSplitPos((byte) b);
			if (splitPos == -1) {
				out.write(b);
				rule.update((byte) b);
			} else if (splitPos == 1) {
				out.write(b);
				rule.update((byte) b);
				next();
			} else {
				throw new IOException("Cannot split one byte");
			}
		} else {
			out.write(b);
		}
	}

	private void next() throws IOException {
		if (footer != null)
			out.write(footer);
		out.flush();
		out.close();
		out = null;
		index++;
		outMode = OutMode.START;
	}

	private boolean shouldCount() throws IOException {
		switch (outMode) {
		case START:
			out = factory.createOutputStream(index);
			if (header != null)
				out.write(header);
			outMode = OutMode.NORMAL;
			return true;
		case NORMAL:
			return true;
		case FOOTER:
		case HEADER:
			return false;
		case CLOSED:
		default:
			throw new IOException("Stream closed");
		}
	}

	/**
	 * Start the header block. After calling this method, all output will be
	 * added to the header of each file, until {@link #finishBlock()} is called.
	 */
	public void startHeaderBlock() {
		if (outMode != OutMode.START || header != null)
			throw new IllegalStateException();
		outMode = OutMode.HEADER;
		out = new ByteArrayOutputStream();
	}

	/**
	 * Start the footer block. After calling this method, all output will be
	 * added to the footer of each file, until {@link #finishBlock()} is called.
	 */
	public void startFooterBlock() {
		if (outMode != OutMode.START || footer != null)
			throw new IllegalStateException();
		outMode = OutMode.FOOTER;
		out = new ByteArrayOutputStream();
	}

	/**
	 * Finish a header or footer block.
	 * 
	 * @see #startHeaderBlock()
	 * @see #startFooterBlock()
	 */
	public void finishBlock() throws IOException {
		if (outMode == OutMode.HEADER) {
			out.flush();
			out.close();
			header = ((ByteArrayOutputStream) out).toByteArray();
		} else if (outMode == OutMode.FOOTER) {
			out.flush();
			out.close();
			footer = ((ByteArrayOutputStream) out).toByteArray();
		} else {
			throw new IllegalStateException();
		}
		outMode = OutMode.START;
		out = null;
	}

	@Override
	public void flush() throws IOException {
		if (out != null)
			out.flush();
	}

	@Override
	public void close() throws IOException {
		flush();
		if (out != null) {
			if (footer != null)
				out.write(footer);
			out.flush();
			out.close();
		}
		outMode = OutMode.CLOSED;
	}

	/**
	 * Factory used to create new streams whenever the stream has to be split.
	 */
	public static interface StreamFactory {

		/**
		 * Create a new stream.
		 * 
		 * @param index
		 *            Zero-based index of the stream
		 */
		public OutputStream createOutputStream(int index) throws IOException;
	}

	private static enum OutMode {
		START, HEADER, FOOTER, NORMAL, CLOSED;
	}
}
